import pytest

from metalpy.mexin.injectors import after, before, terminate_with, modify_params


def func(x):
    print(x)
    return 'Calling from func'


def test_decorator(capsys):
    @before(func)
    def before_test(x):
        print(f"before {x}")

    @after(func)
    def after_test(x):
        print(f"after {x}")

    assert func('test') == 'Calling from func'
    assert capsys.readouterr().out == 'before test\ntest\nafter test\n'

    @before(func)
    def modify_ret_value_test(x):
        print(f"before before {x}")
        return terminate_with('Calling from modify_ret_value_test')

    assert func('test') == 'Calling from modify_ret_value_test'
    assert capsys.readouterr().out == 'before before test\n'


def test_decorator2(capsys):
    import some_injection
    from some_module import some_func, Clazz, DummyClass

    assert some_func(1) == 1
    assert capsys.readouterr().out == 'before some_func\nafter some_func\n'

    x = Clazz()
    assert capsys.readouterr().out == 'before Clazz.__init__\n'

    assert x.extended_method(1) == 1

    y = DummyClass()
    assert y == 1


def test_mixin(capsys):
    from metalpy.mexin import Mixin
    from metalpy.mexin.mixed import Mixed
    from metalpy.mexin.patch_context import patched
    from some_module import ValueHolder

    class SelfIncrement(Mixin):
        def __init__(self, this: ValueHolder):
            super().__init__(this)

        def increase(self, this):
            this.x += 1

    with patched(Mixed(ValueHolder).mix(SelfIncrement)):
        dummy = ValueHolder(1)

        assert dummy.x == 1
        dummy.increase()
        assert dummy.x == 2

    from metalpy.mexin import Patch

    class DummyPatch(Patch):
        def __init__(self, x):
            super().__init__()
            self.x = x

        def apply(self):
            def reassign(this, x):
                this.x = self.x

            self.add_injector(after(ValueHolder.__init__), reassign)
            self.add_mixin(ValueHolder, SelfIncrement)

    with patched(DummyPatch(x=2)):
        dummy = ValueHolder(x=1)
        dummy.increase()
        assert dummy.x == 3

    dummy = ValueHolder(x=1)
    assert dummy.x == 1
    with pytest.raises(AttributeError):
        dummy.increase()
    with pytest.raises(AttributeError):
        _ = dummy.mixins


class DummyClassWithComplicateFunction:
    def func(self, a, b, c='c', *args, d='d', e='e', **kwargs):
        return a, b, c, d, e, args, kwargs


def test_parameters_update():
    inst = DummyClassWithComplicateFunction()

    @after(inst.func)
    def after_func2(this, a, b, c, *args, d, e, **kwargs):
        return a + 1, b, c, d, e, args, kwargs

    @before(inst.func)
    def before_func2(this, a, b, c, *args, d, e, **kwargs):
        return modify_params({
            0: 0,
            3: 'varargs0',
            4: 'varargs1',
        }, b='new b')

    assert inst.func('a', 'b', 'c', '?', '!', d='d', e='e', f='f') == \
           (0 + 1, 'new b', 'c', 'd', 'e', ('varargs0', 'varargs1'), {'f': 'f'})



from __future__ import annotations

import re
import warnings
from typing import Union, Callable, TYPE_CHECKING

import numpy as np

from metalpy.carto.utils.crs import check_crs, CRSLike
from metalpy.utils.bounds import Bounds
from metalpy.utils.numpy import FixedShapeNDArray
from metalpy.utils.string import format_string_list

CRSQuery = Union[str, re.Pattern, Callable]

if TYPE_CHECKING:
    from pyproj import CRS, Transformer


class Coordinates(FixedShapeNDArray):
    """用于表示一个带有坐标系信息的坐标，支持进行便捷坐标转换

    Examples
    --------
    >>> coord = Coordinates([[139.6817, 35.6794], [139.6917, 35.6894]], crs='WGS 84')
    >>> coord.warp(query=Coordinates.SearchUTM)
    Coordinates([[ 380702.31913457, 3949190.73587477],
                [ 381622.08225664, 3950287.81606255]], crs='WGS 84 / UTM zone 54N')
    """
    WGS_84 = 'WGS 84'
    SearchUTM = '__SearchUTM'

    def __new__(cls, arr, crs: CRSLike = WGS_84):
        return np.asarray(arr).view(Coordinates).with_crs(crs)

    def __array_finalize__(self, obj, **kwargs):
        self._init(getattr(obj, 'crs', None))

    def __array_ufunc__(self, *args, **kwargs):
        ret = super().__array_ufunc__(*args, **kwargs)
        if isinstance(ret, Coordinates):
            ret.with_crs(self.crs)

        return ret

    def __getitem__(self, index):
        # TODO: 决定什么时候切片的返回值可以继承Coordinates的属性
        return self.to_numpy()[index]

    def __setitem__(self, index, val):
        self.to_numpy()[index] = val

    def __repr__(self):
        ret = super().__repr__()
        return ret.replace(')', f', crs={repr(self.crs.name)})')

    def _init(self, crs=None):
        if crs is None:
            crs = 'WGS 84'

        self.crs = crs
        assert self.ndim in (1, 2) and self.shape[-1] in (2, 3), 'Coordinates supports only 2D or 3D points array.'

    @property
    def crs(self) -> 'CRS':
        return getattr(self, '_crs', None)

    @crs.setter
    def crs(self, crs_or_name: CRSLike):
        setattr(self, '_crs', check_crs(crs_or_name))

    def to_numpy(self):
        return self.view(np.ndarray)

    @property
    def single(self):
        return self.ndim == 1

    @property
    def x(self):
        if self.single:
            return self[0]
        else:
            return self[:, 0]

    @x.setter
    def x(self, val):
        if self.single:
            self[0] = val
        else:
            self[:, 0] = val

    @property
    def y(self):
        if self.single:
            return self[1]
        else:
            return self[:, 1]

    @y.setter
    def y(self, val):
        if self.single:
            self[1] = val
        else:
            self[:, 1] = val

    @property
    def z(self):
        if self.single:
            return self[2]
        else:
            return self[:, 2]

    @z.setter
    def z(self, val):
        if self.single:
            self[2] = val
        else:
            self[:, 2] = val

    @property
    def bounds(self) -> Bounds:
        if self.single:
            return np.asarray(self).repeat(2).view(Bounds)
        else:
            return np.c_[self.min(axis=0), self.max(axis=0)].ravel().view(Bounds)

    @staticmethod
    def warn_invalid_modification():
        warnings.warn('Modifying dims of Coordinates is not allowed.')

    def reshape(self, *_, **__):
        Coordinates.warn_invalid_modification()

    def with_crs(self, crs):
        self.crs = crs
        return self

    def warp(self,
             crs: CRSLike | None = None,
             query: CRSQuery | None = None,
             inplace=False) -> 'Coordinates':
        """将所有坐标转换到指定的另一坐标系下

        Parameters
        ----------
        crs
            新的坐标系
        query
            按指定模式搜索坐标系
        inplace
            指定转换是否就地操作

        Returns
        -------
        ret
            新坐标系下的坐标集合

        Notes
        -----
        query支持以下方式进行搜索:

        * 指定Coordinates.SearchUTM以搜索UTM分区

        * 指定字符串`CGCS2000 / 3-degree Gauss-Kruger zone`以包含指定串的坐标系

        * 指定正则模式`re.compile('CGCS2000.*3-degree.*Gauss-Kruger.*')`以搜索匹配的坐标系

        * 指定函数`lambda crs: crs.code == 3857`以搜索匹配的坐标系
        """
        from pyproj import CRS, Transformer

        if isinstance(query, CRS):
            crs = query
            query = None

        if crs is not None:
            from pyproj.exceptions import CRSError

            try:
                crs = check_crs(crs)
            except CRSError:
                warnings.warn(f'Unknown CRS `{crs}`. Did you mean `query={repr(crs)}`?'
                              f' Trying to search for related CRS.')
                query = crs
                crs = None

        if crs is None:
            from pyproj.aoi import AreaOfInterest
            from pyproj.database import query_utm_crs_info, query_crs_info

            assert query is not None, 'Either `crs` or `query` must be specified.'

            bounds = self.bounds
            bounds2d = bounds[:4]

            if self.crs == Coordinates.WGS_84:
                wgs84_bounds = bounds2d
            else:
                to_wgs84 = Transformer.from_crs(self.crs, Coordinates.WGS_84, always_xy=True)
                wgs84_bounds = np.c_[
                    to_wgs84.transform(*bounds2d[::2]),
                    to_wgs84.transform(*bounds2d[1::2])
                ].ravel()

            l, r, b, t = wgs84_bounds
            area_of_interest = AreaOfInterest(
                west_lon_degree=l,
                east_lon_degree=r,
                south_lat_degree=b,
                north_lat_degree=t
            )

            if query is Coordinates.SearchUTM:
                crs_list = query_utm_crs_info(
                    datum_name='WGS 84',
                    area_of_interest=area_of_interest,
                )
            else:
                crs_list = query_crs_info(area_of_interest=area_of_interest)
                if isinstance(query, re.Pattern):
                    crs_list = [crs for crs in crs_list if query.match(crs.name)]
                elif isinstance(query, str):
                    # e.g. CGCS2000 / 3-degree Gauss-Kruger zone
                    crs_list = [crs for crs in crs_list if query in crs.name]
                elif isinstance(query, Callable):
                    crs_list = [crs for crs in crs_list if query(crs)]
                else:
                    crs_list = [crs for crs in crs_list if query == crs]

            assert len(crs_list) == 1, \
                f'Multiple CRS-s found by `{query}`: ' + Coordinates.format_crs_list_str(crs_list)

            crs = CRS.from_epsg(crs_list[0].code)

        transform = Transformer.from_crs(self.crs, crs, always_xy=True)

        if inplace:
            target = self
        else:
            target = self.copy()

        target.x, target.y = transform.transform(target.x, target.y)
        target.crs = crs

        return target

    @staticmethod
    def format_crs_list_str(crs_list):
        return format_string_list([crs.name for crs in crs_list], multiline=True)



from typing import Iterable

from metalpy.utils.string import replace_batch

windows_forbidden_chars = r'\/:*?"<>|'


def to_url_escape_sequence(c):
    if c == '%':
        return '%%'
    else:
        return '%{:02X}'.format(ord(c))


def get_char_escape_mapping(chars: Iterable[str], inverse=False):
    if not inverse:
        return {c: to_url_escape_sequence(c) for c in set(chars) | set('%')}
    else:
        return {to_url_escape_sequence(c): c for c in set(chars) | set('%')}


def get_char_escape_trans(chars: Iterable[str], inverse=False):
    return str.maketrans(get_char_escape_mapping(chars, inverse=inverse))


def pathencode(path, forbidden=windows_forbidden_chars):
    path = replace_batch(path, get_char_escape_mapping(forbidden))
    return path


def pathdecode(path, forbidden=windows_forbidden_chars):
    path = replace_batch(path, get_char_escape_mapping(forbidden, inverse=True))
    return path



import enum
from typing import Union


class ArrayType(enum.Enum):
    numpy = 'numpy'
    pandas = 'pandas'

    @staticmethod
    def test(arr):
        module_name = type(arr).__module__
        for atype in ArrayType:
            if module_name.startswith(atype.value):
                return atype


ArrayModuleType = Union[ArrayType, str]


class OperatorDispatcher:
    def __init__(self, op_name: str):
        self.op_name: op_name = op_name
        self.ops: dict[str, callable] = {}

    def register(self, module_name: ArrayModuleType):
        def wrapper(func):
            if isinstance(module_name, ArrayType):
                name = module_name.value
            else:
                name = module_name
            self.ops[name] = func
            return func

        return wrapper

    def alias(self, module_name, alias):
        self.register(module_name)(
            OperatorDispatcher.make_alias_wrapper(module_name, alias)
        )

        return self

    def aliases(self, aliases: dict[ArrayType, str]):
        for module_name, alias in aliases.items():
            self.alias(module_name, alias)

        return self

    register_aliases = aliases

    def dispatch(self, arr_type):
        module_name = arr_type.__module__
        for k, func in self.ops.items():
            if module_name.startswith(k):
                return func

        raise ValueError(f'Unsupported operator {self.op_name} for `{arr_type.__name__}`.')

    @staticmethod
    def make_alias_wrapper(module_name, op_name):
        def wrapper(*args, **kwargs):
            assert isinstance(module_name, ArrayType), \
                'For safety concern, aliases are only allowed in registered packages.'
            func = OperatorDispatcher.import_from_module(module_name.value, op_name)
            return func(*args, **kwargs)

        return wrapper

    @staticmethod
    def import_from_module(module_name, op_name):
        module = __import__(module_name, fromlist=(op_name,))
        return getattr(module, op_name)



from __future__ import annotations

import contextlib
import re
from abc import ABC, abstractmethod
from collections.abc import Collection
from typing import Callable, Any

from .parallel_progress import ParallelProgress
from .task_allocator import TaskAllocator, BoundTaskAllocator, ShuffleChoice
from .utils import structured_traverse
from .worker import Worker
from .worker_context import WorkerContext, BoundWorkerContext, IndexedBoundWorkerContext


class Executor(ABC):
    def __init__(self, event_interval=0.5):
        self.event_interval = event_interval
        self._running = False
        self._event_monitor_thread = None
        self._event_handlers: dict[str, list[callable]] = {}

        self._anonymous_event_counter = 0

    def submit(self, func, *args, worker=None, workers=None, **kwargs):
        """发布任务

        Parameters
        ----------
        func
            待执行的函数
        worker, workers
            指定需要绑定的worker
        args, kwargs
            函数参数

        Returns
        -------
        future
            所分配任务的future
        """
        workers = check_workers(worker, workers)
        return self.do_submit(func, workers=workers, *args, **kwargs)

    def distribute(self, func, *args, **kwargs):
        """将任务分派到所有worker上

        对 `args` 和 `kwargs` 中的 `TaskAllocator` 会按权重分配到worker上。

        Parameters
        ----------
        func
            待执行的函数
        args, kwargs
            函数参数

        Returns
        -------
        futures
            所分配任务的futures，长度与执行器实例的worker数相同
        """
        ret = []
        for w in self.get_workers():
            def check_param(p):
                if isinstance(p, TaskAllocator):
                    return p.assign(w)
                else:
                    return p

            ret.append(self.submit(
                func,
                *structured_traverse(list(args), check_param),
                **structured_traverse(kwargs, check_param),
                worker=w
            ))
        return ret

    def map(self, func, *iterables, worker=None, workers=None, chunksize=None):
        workers = check_workers(worker, workers)

        futures = []
        n_workers = len(workers)

        for i, task in enumerate(zip(*iterables)):
            if workers is not None:
                w = [workers[(i // chunksize) % n_workers]]
            else:
                w = None
            futures.append(self.do_submit(func, *task, workers=w))

        return futures

    def extract_by_name(self, pat: str | re.Pattern | Callable):
        """筛选符合条件的worker构成子执行器

        Parameters
        ----------
        pat
            模式，支持通过字符串包含worker名/正则匹配worker名/函数谓词判定/重载`==`运算符等方式进行

        Returns
        -------
        sub_executor
            提取的worker构造的新的执行器
        """
        from .sub_executor import SubExecutor

        if isinstance(pat, str):
            workers = [w for w in self.get_workers() if pat in w.get_name()]
        elif isinstance(pat, re.Pattern):
            workers = [w for w in self.get_workers() if pat.match(w.get_name())]
        elif isinstance(pat, Callable):
            workers = [w for w in self.get_workers() if pat(w)]
        else:
            workers = [w for w in self.get_workers() if pat == w]

        return SubExecutor(parent=self, workers=workers)

    @abstractmethod
    def do_submit(self, func, *args, workers=None, **kwargs):
        pass

    def arrange(self, data, shuffle: ShuffleChoice = False) -> BoundTaskAllocator:
        """将给定序列数据按当前worker的权重分割

        Parameters
        ----------
        data
            序列数据
        shuffle
            指示是否打乱数据

        Returns
        -------
        allocator
            数据分配器，遍历即可得到按权重分配的数据
        """
        return BoundTaskAllocator.arrange(
            data,
            weights=self.get_workers(),
            shuffle=shuffle
        )

    def arrange_many(self, *data_list, shuffle: ShuffleChoice = False) -> BoundTaskAllocator:
        """将给定多个序列数据按当前worker的权重分割，并类似zip(*)进行组合

        Parameters
        ----------
        data_list
            多个序列数据
        shuffle
            指示是否打乱数据

        Returns
        -------
        allocator
            数据组分配器，遍历即可得到按权重分配的数据组

        Examples
        --------
        >>> import numpy as np
        >>> from metalpy.mepa import ThreadExecutor
        >>> executor = ThreadExecutor(3)
        >>> tasks = executor.arrange_many(np.arange(7), np.arange(7, 14))
        >>> print(list(tasks))
        <<< [(array([0, 1, 2]), array([7, 8, 9])),
        <<<  (array([3, 4, 5]), array([10, 11, 12])),
        <<<  (array([6]), array([13]))]
        """
        return BoundTaskAllocator.arrange_many(
            *data_list,
            weights=self.get_workers(),
            shuffle=shuffle
        )

    def progress(self, total=0, **kwargs) -> ParallelProgress:
        """构造并行进度条，用于快捷实现分布式进度展示

        Parameters
        ----------
        total
            总任务量
        kwargs
            其它需要传递给 `tqdm` 进度条构造函数的参数

        Returns
        -------
        parallel_progress
            主机侧并行进度条，可以通过函数捕获或传参，传递给worker使用

        See Also
        --------
        :class:`ParallelProgress` : 并行进度条实现
        """
        return ParallelProgress(self, total=total, **kwargs)

    @abstractmethod
    def get_workers(self) -> Collection[Worker]:
        pass

    def get_n_workers(self):
        """返回当前集群的总worker数
        """
        return len(self.get_workers())

    def get_total_weights(self):
        """返回当前集群所有worker的权重和
        """
        return sum([w.get_weight() for w in self.get_workers()])

    @abstractmethod
    def is_local(self):
        return True

    def gather(self, futures):
        """收集futures的结果

        Parameters
        ----------
        futures
            待收集的futures，可以为单个future也可以是future的Iterable或dict

        Returns
        -------
        ret
            以原本结构构造的future结果

        Examples
        --------
        >>> from operator import add
        >>> from metalpy.mepa import LinearExecutor
        >>> executor = LinearExecutor()
        >>> f = executor.submit(add, 1, 2)
        >>> executor.gather(f)
        <<< 3
        >>> executor.gather([f, [f], f])  # 支持列表
        <<< [3, [3], 3]
        >>> executor.gather({'First': f, 'Second': [f, [f]]})  # 支持字典
        <<< {'First': 3, 'Second': [3, [3]]}
        """
        with self.monitor_events():
            return self._gather(futures)

    def gather_single(self, future):
        """收集单个的结果

        Parameters
        ----------
        future
            待收集的future

        Returns
        -------
        ret
            future的结果
        """
        with self.monitor_events():
            return self._gather_single(future)

    def submit_and_gather(self, func, *args, worker=None, workers=None, **kwargs):
        fut = self.submit(func, *args, worker=worker, workers=workers, **kwargs)
        return self.gather(fut)

    def distribute_and_gather(self, func, *args, **kwargs):
        futures = self.distribute(func, *args, **kwargs)
        return self.gather(futures)

    def _gather(self, futures):
        """收集futures的结果的实现逻辑

        Parameters
        ----------
        futures
            待收集的futures，可以为单个future也可以是future的Iterable或dict

        Returns
        -------
        ret
            以原本结构构造的future结果

        Notes
        -----
            子类需从_gather和_gather_single二选一实现
            子类实现_gather时建议实现structured_traverse逻辑
        """
        return structured_traverse(futures, lambda f: self._gather_single(f))

    def _gather_single(self, future):
        """收集单个的结果的实现逻辑

        Parameters
        ----------
        future
            待收集的future

        Returns
        -------
        ret
            future的结果

        Notes
        -----
            子类需从_gather和_gather_single二选一实现
        """
        return self._gather(future)

    def scatter(self, data):
        """
        用于预先分发大规模数据（一般指内存占用大于1MB）

        Parameters
        ----------
        data
            大规模数据

        See Also
        --------
        distributed.Client.scatter
        """
        return data

    @abstractmethod
    def get_worker_context(self) -> WorkerContext:
        """获取当前Executor绑定的WorkerContext实例

        Returns
        -------
        context
            返回当前Executor绑定的WorkerContext实例，如果没有，则创建

        Notes
        -----
        一般情况下，所有事件共享一个WorkerContext实例，发布消息时由Executor完成消息分发
        """
        pass

    @abstractmethod
    def receive_events(self, timeout):
        pass

    def on(self, event, callback) -> IndexedBoundWorkerContext:
        """绑定子worker发出的消息事件回调。

        Executor会在单独的监听线程中获取消息事件并调用回调函数。

        Parameters
        ----------
        event
            监听的事件标记
        callback
            事件回调函数

        Returns
        -------
        context
            返回绑定到该事件的WorkerContext，可以直接通过fire调用到该回调函数，不需要指定事件
        """
        callbacks = self._event_handlers.setdefault(event, [])
        callbacks.append(callback)

        return self.get_worker_context().bind(event, len(callbacks))

    def register(self, callback) -> BoundWorkerContext:
        """绑定一个匿名消息的事件回调。

        Parameters
        ----------
        callback
            事件回调函数

        Returns
        -------
        context
            返回绑定到该事件的WorkerContext，可以直接通过fire调用到该回调函数，不需要指定事件

        See Also
        --------
        Executor.on : 会自动生成一个专有名字然后通过on绑定
        """
        event = _Anonymous(self._anonymous_event_counter)
        self._anonymous_event_counter += 1

        return self.on(event, callback)

    def off(self, event: BoundWorkerContext | Any, callback=None):
        """移除子worker绑定的消息事件回调。

        Parameters
        ----------
        event
            监听的事件标记
        callback
            事件回调函数，如果不指定，则指定事件下的所有回调函数
        """
        if isinstance(event, BoundWorkerContext):
            event_key = event.event
            idx = None
            if isinstance(event, IndexedBoundWorkerContext):
                idx = event.idx
        else:
            event_key = event
            if callback is not None:
                # 指示需要进行搜索
                idx = callback
            else:
                # 指示全部移除
                idx = None

        if idx is None:
            # 全部移除
            self._event_handlers.pop(event)
        else:
            callbacks = self._event_handlers.get(event_key, None)
            if callbacks is not None:
                if not isinstance(idx, int):
                    idx = callbacks.index(callback)
                callbacks[idx] = None
                pop_trailing_nones(callbacks)
                if len(callbacks) == 0:
                    self._event_handlers.pop(event)

    def dispatch_event(self, event, args, kwargs):
        listeners = self._event_handlers.get(event, None)
        if listeners is not None:
            for listener in listeners:
                if listener is None:
                    continue
                listener(*args, **kwargs)

    def monitor(self):
        for event, msg in self.receive_events(self.event_interval):
            if event is None:
                pass
            else:
                args, kwargs = msg
                self.dispatch_event(event, args, kwargs)

            if not self._running:
                return

    def start_event_monitoring(self, start_thread):
        self._running = True
        if start_thread:
            from threading import Thread
            self._event_monitor_thread = Thread(target=self.monitor)
            self._event_monitor_thread.start()

    def terminate_event_monitoring(self):
        self._running = False

    def check_if_events_thread_are_required(self):
        return len(self._event_handlers) > 0

    @contextlib.contextmanager
    def monitor_events(self):
        """启动worker消息事件的监听线程

        Notes
        -----
        `gather`和`gather_single`会自动调用该函数启动监听线程、
        """
        if not self._running:
            try:
                self.start_event_monitoring(self.check_if_events_thread_are_required())
                yield
            finally:
                self.terminate_event_monitoring()
        else:
            try:
                yield
            finally:
                pass

    def shutdown(self, wait=True):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown(wait=True)


class _Anonymous:
    __slots__ = ['i']

    def __init__(self, i):
        self.i = i

    def __hash__(self):
        return hash(f'__A{self.i}')

    def __eq__(self, other):
        return isinstance(other, _Anonymous) and self.i == other.i


def check_workers(worker, workers):
    if workers is None:
        if worker is None:
            workers = None
        else:
            workers = [worker]
    else:
        workers = list(workers)

    return workers


def pop_trailing_nones(elems):
    while elems[-1] is None:
        elems.pop()



from __future__ import annotations

from abc import ABC, abstractmethod

from .injectors import RecoverableInjector
from .mixin import Mixin


class Patch(ABC):
    Priority = DefaultPriority = 225

    def __init__(self):
        """用于在全局上下文中劫持与注入python-based方法，继承类必须重写apply方法

        Notes
        -----
            所有方法必须保持可重入，即多次调用apply方法，必须保证结果一致
            rollback会消除commit的影响
            post_rollback会消除pre_apply和apply的影响


            执行时PatchContext会按如下顺序调用方法：

            **bind_context**: 绑定上下文

            **pre_apply**: 在apply之前调用，用于初始化一些变量

            **apply**: 在这里添加patch，子类必须实现

            **commit**: 在这里进行patch的应用，如果要改写，应该需要调用基类方法

            **rollback**: 在这里进行一些rollback操作

            **post_rollback**: 在rollback之后调用，用于清理一些变量

            **unbind_context**: 解绑上下文

        See Also
        --------
            PatchContext: patch上下文
        """
        self.recoverable_injs: list[tuple[RecoverableInjector, tuple, dict]] | None = None
        self.mixins: dict[type, list[tuple[Mixin, tuple, dict]]] | None = None
        self._context = None

    def get_mixed_classes(self):
        return self.mixins.keys()

    def pre_apply(self):
        self.recoverable_injs = []
        self.mixins = {}

    @abstractmethod
    def apply(self):
        pass

    def commit(self):
        for inj, args, kwargs in self.__get_injs():
            inj(*args, **kwargs)

    def rollback(self):
        for inj, args, kwargs in self.__get_injs(reverse=True):
            inj.rollback()

    def post_rollback(self):
        # 防止其中的lambda函数和类指针被pickle
        self.recoverable_injs = None
        self.mixins = None

    def __get_injs(self, reverse=False):
        if reverse:
            return reversed(self.recoverable_injs)
        else:
            return self.recoverable_injs

    @property
    def priority(self):
        # -1 专属于Mixed的优先级，用于注入Mixin系统
        # TODO: 可能可以改为require/after类型的依赖图？
        return type(self).Priority

    @property
    def context(self):
        return self._context

    def bind_context(self, ctx):
        self._context = ctx

    def unbind_context(self):
        self._context = None

    def add_injector(self, inj: RecoverableInjector, *args, **kwargs):
        """追加injector，执行全局的替换或劫持等操作

        Parameters
        ----------
        inj
            待添加的injector对象
        args
            注入操作（injector.__call__）的位置参数
        kwargs
            注入操作（injector.__call__）的关键字参数

        See Also
        --------
            PatchContext.apply: 应用injector
        """
        self.recoverable_injs.append((inj, args, kwargs))

    def add_mixin(self, target_type, mixin_type: [Mixin], *args, **kwargs):
        """指定向target_type添加局限于类实例的Mixin，在作例用范围内创建的所有target_type实例都会对应创建并绑定一份该Mixin实例

        Parameters
        ----------
        target_type
            需要添加Mixin的目标类
        mixin_type
            待添加的Mixin类
        args
            Mixin类的位置参数
        kwargs
            Mixin类的关键字参数

        Notes
        -----
            Patch本身不会进行任何关于Mixin的操作，Mixin的创建与绑定由PatchContext完成

        See Also
        --------
            PatchContext.apply: 执行MixinManager的创建与绑定
            Mixed.with_mixin: 执行Mixin的定义
            Mixed.apply_mixin or MixinManager.add: 执行Mixin的创建与绑定
        """
        self.mixins.setdefault(target_type, []).append((mixin_type, args, kwargs))



from enum import Enum
from typing import Union, Callable


class MixMode(Enum):
    @staticmethod
    def dispatch(mode):
        if isinstance(mode, MixMode):
            name = str(mode.value)
            return getattr(MixMode, name)
        elif callable(mode):
            return mode
        else:
            raise ValueError(f'Mix mode must be either MixMode or function, got {type(mode)} instead.')

    @staticmethod
    def override(prev_layer, current_layer):
        return current_layer

    @staticmethod
    def keep_original(prev_layer, current_layer):
        return prev_layer

    @staticmethod
    def max(prev_layer, current_layer):
        mask = prev_layer < current_layer
        prev_layer[mask] = current_layer[mask]
        return prev_layer

    @staticmethod
    def min(prev_layer, current_layer):
        mask = prev_layer > current_layer
        prev_layer[mask] = current_layer[mask]
        return prev_layer

    @staticmethod
    def farther_from_zero(prev_layer, current_layer):
        mask = abs(prev_layer) < abs(current_layer)
        prev_layer[mask] = current_layer[mask]
        return prev_layer

    @staticmethod
    def closer_to_zero(prev_layer, current_layer):
        mask = abs(prev_layer) > abs(current_layer)
        prev_layer[mask] = current_layer[mask]
        return prev_layer

    @staticmethod
    def mean(prev_layer, current_layer):
        return (prev_layer + current_layer) / 2

    Override = override.__func__.__name__  # 直接覆盖
    KeepOriginal = keep_original.__func__.__name__  # 保留原值
    Max = max.__func__.__name__  # 取最大值 / 正数意义下的并集 / 负数意义下的交集
    Min = min.__func__.__name__  # 取最小值 / 正数意义下的交集 / 负数意义下的并集
    AbsMax = FartherFromZero = farther_from_zero.__func__.__name__  # 取绝对值意义下最大值 / 更远离原点的值 / 同符号意义下的并集
    AbsMin = CloserToZero = closer_to_zero.__func__.__name__  # 取绝对值意义下最小值 / 更接近原点的值 / 同符号意义下的交集
    Mean = mean.__func__.__name__  # 取平均值


Mixer = Union[MixMode, Callable]


def dhashable_mixer(mixer: Mixer):
    """返回mixer的哈希值

    Parameters
    ----------
    mixer
        混合器名或函数
    """
    if isinstance(mixer, MixMode):
        return mixer.value
    else:
        import cloudpickle
        return cloudpickle.dumps(mixer)



import copy


class FunctionTermination:
    def __init__(self, *rets):
        """指示提前终止函数，目前仅在before注解中使用，指示提前终止函数
        """
        if len(rets) == 0:
            self.ret = None
        elif len(rets) == 1:
            self.ret = rets[0]
        else:
            self.ret = rets


def terminate() -> FunctionTermination:
    """指示无返回值提前终止函数执行

    Returns
    -------
    ret
        由Before注解接收并执行函数终止操作
    """
    return FunctionTermination()


def terminate_with(*rets) -> FunctionTermination:
    """指示以指定的返回值提前终止函数

    Parameters
    ----------
    rets
        指定函数终止的返回值

    Returns
    -------
    ret
        由Before注解接收并执行函数终止操作，然后返回指定的返回值
    """
    return FunctionTermination(*rets)


class ParameterModification:
    def __init__(self, new_args):
        """指示修改函数参数，目前仅在before注解中使用，指示修改函数参数
        """
        self.new_args = new_args


def modify_params(new_args=None, **new_kwargs) -> ParameterModification:
    """指示修改函数参数

    Parameters
    ----------
    new_args
        以字典形式传入的参数修改指示，int作为key用于修改指定位置的位置参数，str作为key用于修改关键字参数或以关键字形式指定的位置参数
    new_kwargs
        以关键字形式传入的便捷参数修改指示，用于修改关键字参数或以关键字形式指定的位置参数

    Returns
    -------
    ret
        由Before注解接收并执行参数修改
    """
    if new_args is None:
        new_args = {}
    else:
        new_args = copy.copy(new_args)
    new_args.update(new_kwargs)
    return ParameterModification(new_args)



import pyvista as pv

from metalpy.carto.coords import Coordinates
from metalpy.utils.model import pv_ufunc_apply, DataSetLike


def warp_dataset(dataset: DataSetLike, src_crs, *, crs=None, query=None, inplace=False, return_crs=False):
    """转换数据集的坐标系

    Parameters
    ----------
    dataset
        待转换坐标系
    src_crs
        源坐标系
    crs
        目标坐标系
    query
        按指定模式搜索坐标系
    inplace
        指定转换是否就地操作
    return_crs
        指示时候返回采用的新坐标系，常用与配合`query`，获取查询到的坐标系信息

    Returns
    -------
    result
        转换坐标系后的数据集，如果指定了`return_crs=True`，则额外返回转换后的坐标系信息

    See Also
    --------
    :py:class:`metalpy.carto.coords.Coordinates` : 用于表示与转换坐标
    """
    dest_crs = [None]

    def wrapper(d):
        if dest_crs[0] is None:
            _, dest_crs[0] = _warp_dataset(d, src_crs, crs=crs, query=query, return_crs=True)
        else:
            _ = _warp_dataset(d, src_crs, crs=dest_crs[0])

    warped = pv_ufunc_apply(
        dataset, wrapper,
        inplace=inplace
    )

    if return_crs:
        return warped, dest_crs[0]
    else:
        return warped


def _warp_dataset(dataset: pv.DataSet, src_crs, *, crs=None, query=None, return_crs=False):
    geo_points: Coordinates = dataset.points.view(Coordinates).with_crs(src_crs)

    dest_crs = None
    if crs is not None or query is not None:
        geo_points.warp(crs=crs, query=query, inplace=True)
        try:
            dataset.points = geo_points
        except AttributeError:
            # 例如RectilinearGrid这类参数化对象无法直接设置模型点位置
            # 转换为UnstructuredGrid再进行转换
            dataset = dataset.cast_to_unstructured_grid()
            dataset.points = geo_points

        if return_crs:
            dest_crs = geo_points.crs

    if return_crs:
        return dataset, dest_crs
    else:
        return dataset



from random import random

import numpy as np

from metalpy.scab.utils.misc import define_inducing_field as new_define_inducing_field


def define_inducing_field(s, i, d):
    """
    Deprecated. Use metalpy.scab.utils.misc.define_inducing_field instead
    """
    return new_define_inducing_field(s, i, d)


def plot_opaque_cube(ax, x=10, y=20, z=30, dx=40, dy=50, dz=60, alpha=1, **kwargs):
    xx = np.linspace(x, x + dx, 2)
    yy = np.linspace(y, y + dy, 2)
    zz = np.linspace(z, z + dz, 2)

    xx2, yy2 = np.meshgrid(xx, yy)

    ax.plot_surface(xx2, yy2, np.full_like(xx2, z), alpha=alpha, **kwargs)
    ax.plot_surface(xx2, yy2, np.full_like(xx2, z + dz), alpha=alpha, **kwargs)

    yy2, zz2 = np.meshgrid(yy, zz)
    ax.plot_surface(np.full_like(yy2, x), yy2, zz2, alpha=alpha, **kwargs)
    ax.plot_surface(np.full_like(yy2, x + dx), yy2, zz2, alpha=alpha, **kwargs)

    xx2, zz2 = np.meshgrid(xx, zz)
    ax.plot_surface(xx2, np.full_like(yy2, y), zz2, alpha=alpha, **kwargs)
    ax.plot_surface(xx2, np.full_like(yy2, y + dy), zz2, alpha=alpha, **kwargs)


def plot_linear_cube(ax, x, y, z, dx, dy, dz, color='red'):
    xx = [x, x, x + dx, x + dx, x]
    yy = [y, y + dy, y + dy, y, y]
    kwargs = {'alpha': 1, 'color': color}
    ax.plot3D(xx, yy, [z] * 5, **kwargs)
    ax.plot3D(xx, yy, [z + dz] * 5, **kwargs)
    ax.plot3D([x, x], [y, y], [z, z + dz], **kwargs)
    ax.plot3D([x, x], [y + dy, y + dy], [z, z + dz], **kwargs)
    ax.plot3D([x + dx, x + dx], [y + dy, y + dy], [z, z + dz], **kwargs)
    ax.plot3D([x + dx, x + dx], [y, y], [z, z + dz], **kwargs)


def sin2pi(a):
    return np.sin(2 * np.pi * a)


def cos2pi(a):
    return np.cos(2 * np.pi * a)


def rand_between(vmin, vmax):
    return (vmax - vmin) * random() + vmin



import copy
import os

import SimPEG
import blosc2
import numpy as np
from SimPEG.potential_fields import magnetics
from SimPEG.simulation import BaseSimulation
from discretize import TensorMesh
from properties.utils import undefined

from metalpy.mepa import Executor
from metalpy.mexin import LazyClassFactory, PatchContext
from metalpy.mexin.injectors import is_or_is_replacement, reverted
from metalpy.scab.simpeg_patch_context import simpeg_patched
from metalpy.utils.type import not_none_or_default
from .policies import Distributable, NotDistributable
from .utils import reget_class


class DistributedSimulation(LazyClassFactory, BaseSimulation):
    def __init__(self,
                 patch,
                 simulation_class: BaseSimulation,
                 *args,
                 survey: LazyClassFactory,
                 executor: Executor = None,
                 **kwargs):
        """
        Parameters
        ----------
        simulation_class
            正演类，继承自BaseSimulation
        survey
            观测网格和背景条件
        executor
            执行器，默认为None，此时会使用LinearExecutor
        args
            simulation的位置参数
        kwargs
            simulation的关键字参数

        Notes
        -----
            目前只用于magnetics.simulation.Simulation3DIntegral，对于其它Simulation兼容性未知
        """

        super().__init__(simulation_class, *args, **kwargs)

        if not isinstance(survey, LazyClassFactory):
            raise ValueError('Error: survey must be a LazyClassFactory. '
                             'Please check if the patch system is working correctly.')

        source_field = survey.find_param_by_type(LazyClassFactory, remove=True)

        if source_field is None:
            raise ValueError('Error: source field must be a LazyClassFactory. '
                             'Please check if the patch system is working correctly.')

        if not is_or_is_replacement(SimPEG.potential_fields.magnetics.simulation.Simulation3DIntegral,
                                    simulation_class):
            print('Warning: Distributed patch is only tested for magnetics.simulation.Simulation3DIntegral')

        # 针对magnetics.simulation.Simulation3DIntegral实现
        # TODO: 引入来支持其它Simulation，TaskDividePolicy?
        receiver_list = source_field.find_param(lambda k, v: isinstance(k, str) and k.startswith('receiver_'),
                                                remove=True)

        # 和外部共用receiver_list，防止被后文修改
        self.simulation = self.build_simulation(survey, source_field, receiver_list)

        # 复制receiver_list，防止影响到外部代码
        receiver_list = [copy.deepcopy(receiver) for receiver in receiver_list]
        receiver_locations = []
        for receiver in receiver_list:
            receiver_locations.append(receiver.locations)
            receiver.locations = receiver.locations[0]  # 占位符，用来越过receiver的类型检测

        self.simulation_class = simulation_class
        self.survey_factory = survey
        self.source_field = source_field
        self.receiver_list = receiver_list
        self.locations_list = receiver_locations

        self.executor = executor

        self.parallelized_patch = patch

        # 在构造完本地代理类后劫持分派类的store_sensitivities参数，如果为disk，则改为ram，防止在远程端产生不必要的cache和路径冲突
        self.store_sensitivities = self.get('store_sensitivities', default='ram')
        if self.store_sensitivities == 'disk':
            self['store_sensitivities'] = 'ram'
        else:
            self['store_sensitivities'] = self.store_sensitivities
        self.sensitivity_path = self.pop('sensitivity_path', default="./sensitivity/")

    @staticmethod
    def auto_decompress_simulation(sim, ind_future, mesh_future):
        sim['ind_active'] = blosc2.unpack_array2(ind_future)
        sim['mesh'] = mesh_future
        return sim

    @staticmethod
    def auto_decompress_mesh_h(compressed_mesh_defs):
        return [blosc2.unpack_array2(ch) for ch in compressed_mesh_defs]

    def compress_model(self, sim_delegate, model):
        """用于压缩模型，以减少传输量

        主要包含model和simulation里的ind_active

        Parameters
        ----------
        sim_delegate
            仿真类构造器

        model
            模型

        Returns
        -------
        ret
            压缩后的仿真类构造器和模型
        """
        from pympler.asizeof import asizeof

        if asizeof(model) > 1024 * 1024:
            compressed_model = blosc2.pack_array2(model)
            compressed_model = self.executor.scatter(compressed_model)
            model = self.executor.submit(blosc2.unpack_array2, compressed_model)

        if asizeof(sim_delegate) > 1024 * 1024:
            compressed_act_ind = blosc2.pack_array2(sim_delegate['ind_active'])
            compressed_act_ind = self.executor.scatter(compressed_act_ind)
            sim_delegate['ind_active'] = 0

            mesh = sim_delegate.find_param_by_name('mesh', remove=True)
            if isinstance(mesh, TensorMesh):
                if asizeof(mesh.h) > 1024 * 1024:
                    compressed_mesh_h = [blosc2.pack_array2(h) for h in mesh.h]
                    mesh_h = self.executor.submit(self.auto_decompress_mesh_h, compressed_mesh_h)
                else:
                    mesh_h = mesh.h

                origin = mesh.origin
                mesh_future = self.executor.submit(TensorMesh, h=mesh_h, origin=origin)
            else:
                mesh_future = self.executor.scatter(mesh)

            sim_delegate = self.executor.submit(self.auto_decompress_simulation,
                                                sim_delegate,
                                                compressed_act_ind,
                                                mesh_future)

        return sim_delegate, model

    @staticmethod
    def linear_operator_worker(simulation_delegate,
                               survey_delegate,
                               source_field_delegate,
                               receiver_list_container,
                               locations_list,
                               model,
                               patches):
        # model property强制要求为undefined或numpy数组，不可为None
        model = not_none_or_default(model, undefined)

        for receiver in receiver_list_container:
            receiver.locations = locations_list.pop(0)

        with simpeg_patched(*patches):
            survey = survey_delegate.construct(
                source_field=source_field_delegate.construct(
                    receiver_list=receiver_list_container
                ))

            simulation = simulation_delegate.construct(survey=survey)
            simulation.model = model

        return simulation.linear_operator()

    def linear_operator(self, this):
        """LinearSimulation
        """
        n_cells = self.nC
        if getattr(self, "model_type", None) == "vector":
            n_cells *= 3
        if self.store_sensitivities == "disk":
            sens_name = os.path.join(self.sensitivity_path, "sensitivity.npy")
            if os.path.exists(sens_name):
                # do not pull array completely into ram, just need to check the size
                kernel = np.load(sens_name, mmap_mode="r")
                if kernel.shape == (self.survey.nD, n_cells):
                    print(f"Found sensitivity file at {sens_name} with expected shape")
                    kernel = np.asarray(kernel)
                    return kernel

        if PatchContext.lock.locked():
            # 需要等PatchContext退出来让加载的Patch解除对上下文的绑定，从而防止序列化方面的问题
            raise AssertionError('Error: linear_operator must be called outside of PatchContext.')

        simulation_delegate = self.clone()  # 使用clone创建一个lazy构造器
        model = this.model
        if not self.executor.is_local():
            simulation_delegate, model = self.compress_model(simulation_delegate, model)

        futures = []
        receiver_tasks = self.executor.arrange_many(*self.locations_list)
        for dest_worker in self.executor.get_workers():
            # 使用get_patch_policy来判断上下文中应用的patch哪些需要在worker中启用
            patches = [self.get_patch_policy(patch).distribute_to(self.executor, dest_worker)
                       for patch in self.get_patches()]
            patches = [patch for patch in patches if patch is not None]

            future = self.executor.submit(
                self.linear_operator_worker, worker=dest_worker,
                simulation_delegate=simulation_delegate,
                survey_delegate=self.survey_factory,
                source_field_delegate=self.source_field,
                receiver_list_container=self.receiver_list,
                locations_list=receiver_tasks.assign(dest_worker),
                model=model,
                patches=patches
            )
            futures.append(future)

        segments = self.executor.gather(futures)
        kernel = np.concatenate(segments)

        if self.store_sensitivities == "disk":
            print(f"writing sensitivity to {sens_name}")
            os.makedirs(self.sensitivity_path, exist_ok=True)
            np.save(sens_name, kernel)

        return kernel

    def dpred(self, model=None, f=None):
        """BaseSimulation
        """
        return self.simulation.dpred(model, f)

    @property
    def model(self):
        """LinearSimulation
        """
        return self.simulation.model

    @property
    def survey(self):
        """BaseSimulation
        """
        return self.simulation.survey

    def getJtJdiag(self, m, W=None):
        """Simulation3DIntegral / Simulations from electromagnetics
        """
        return self.simulation.getJtJdiag(m, W)

    def fields(self, model):
        """BaseSimulation
        """
        return self.simulation.fields(model)

    def Jtvec(self, m, v, f=None):
        """BaseSimulation
        """
        return self.simulation.Jtvec(m, v, f=f)

    def Jvec(self, m, v, f=None):
        """BaseSimulation
        """
        return self.simulation.Jvec(m, v, f=f)

    def Jtvec_approx(self, m, v, f=None):
        """BaseSimulation
        """
        return self.simulation.Jtvec_approx(m, v, f=f)

    def Jvec_approx(self, m, v, f=None):
        """BaseSimulation
        """
        return self.simulation.Jvec_approx(m, v, f=f)

    def residual(self, m, dobs, f=None):
        """BaseSimulation
        """
        return self.simulation.residual(m, dobs, f=f)

    @property
    def G(self):
        """LinearSimulation
        """
        return self.simulation.G

    @G.setter
    def G(self, value):
        """LinearSimulation
        """
        self.simulation.G = value

    @property
    def nC(self):
        """BasePFSimulation
        """
        return self.simulation.nC

    @staticmethod
    def get_patch_policy(patch):
        if isinstance(patch, Distributable):
            return patch
        else:
            return NotDistributable()

    def get_patches(self):
        return self.parallelized_patch.persisted_context.get_patches()

    def build_simulation(self, survey, source_field, receiver_list):
        r = reget_class(survey.cls), reget_class(source_field.cls),
        with reverted(*r):
            ret = self.construct(
                survey=survey.construct(
                    source_field=source_field.construct(
                        receiver_list=receiver_list)))
        ret.mixins.bind_method(self.linear_operator)
        return ret



from __future__ import annotations

from typing import Union, TYPE_CHECKING

if TYPE_CHECKING:
    from pyproj import CRS


CRSLike = Union['CRS', str]


def check_crs(crs: CRSLike):
    from pyproj import CRS

    if isinstance(crs, str):
        crs = CRS.from_string(crs)

    return crs



from .executor import Executor
from .worker import Worker
from .worker_context import WorkerContext, BoundWorkerContext, IndexedBoundWorkerContext
from .queue_like_worker_context import QueueLikeWorkerContext
from .parallel_progress import ParallelProgress

from .task_allocator import MultiTaskAllocator, SingleTaskAllocator
from .lazy_evaluator import LazyEvaluator

from .linear_executor import LinearExecutor
from .process_executor import ProcessExecutor
from .dask_executor import DaskExecutor
from .thread_executor import ThreadExecutor



import numpy as np

from . import Texture, EmptyTexture


class Specular(Texture):
    def __init__(self, color, shininess):
        super().__init__(np.r_[color, shininess])

    @classmethod
    def extract(cls, dataset, name=None):
        data = cls.extract_field(dataset, name=name)
        if data is None:
            return EmptyTexture()
        return cls(data[:3], data[-1])

    @property
    def color(self):
        return self.texture[:3]

    @property
    def shininess(self):
        return self.texture[-1]

    def apply(self, actor):
        actor.prop.specular_color = self.color
        # PyVista的默认亮度为100.0，但mtl默认为0.0
        actor.prop.specular_power = min(self.shininess + 100, 128)
        actor.prop.specular = 1



import importlib
import inspect
import sys
import warnings
from typing import Optional, Union

from metalpy.utils.dhash import dhash

module_type = type(warnings)


class DottedName:
    def __init__(self, *paths):
        """用于表示用'.'分割的Python对象路径

        Parameters
        ----------
        paths
            一组路径，会被'.'连接
        """
        self.parts = [seg for path in paths for seg in DottedName._split_path(path)]

    def without_prefix(self, *prefixes):
        """获取当前路径移除指定前缀后的路径

        Parameters
        ----------
        prefixes
            路径

        Returns
        -------
        ret
            按顺序依次匹配prefixes中前缀，如果匹配，则移除，否则返回当前结果

        Examples
        --------
        >>> DottedName('x.y.z').without_prefix('x.y', 'z')
        <<< DottedName('')

        >>> DottedName('x.y.z').without_prefix('x', 'z.y')
        <<< DottedName('y.z')  # only 'x' matches the prefix, and is removed
        """
        trimed = 0
        if len(prefixes) == 0:
            trimed = 1
        else:
            for prefix in prefixes:
                prefix = DottedName(prefix)
                matched = True
                for i, rhs in enumerate(prefix.parts):
                    if self.parts[trimed + i] != rhs:
                        matched = False
                        break
                if matched:
                    trimed += len(prefix)
                else:
                    break

        ret = DottedName()
        ret.parts = self.parts[trimed:]
        return ret

    def suffix(self, length=1):
        """获取当前路径中指定长度的后缀

        Parameters
        ----------
        length
            需要获取的后缀长度

        Returns
        -------
        ret
            当前路径的后缀数组，长度小于等于length

        Examples
        --------
        >>> DottedName('x.y.z').suffix()
        <<< DottedName('z')
        """
        ret = DottedName()
        length = min(self.__len__(), length)
        ret.parts = self.parts[-length:]

        return ret

    @property
    def parent(self):
        if len(self) > 1:
            return DottedName(*self.parts[:-1])
        else:
            raise ValueError('Trying to access parent of a top-level DottedName.')

    @property
    def empty(self):
        return self.__len__() == 0

    def __getitem__(self, item):
        return self.parts[item]

    def __setitem__(self, key, value):
        if key < 0:
            key = len(self.parts) + key

        if isinstance(value, str):
            self.__setitem__(key, DottedName(value))
        elif isinstance(value, (list, tuple, DottedName)):
            self.parts = [*self.parts[0:key], *value, *self.parts[key + 1:]]
        else:
            warnings.warn('Result of assigning a non-str object to be part of DottedName may be unexpected.')
            self.__setitem__(key, DottedName(str(value)))

    def __iter__(self):
        for part in self.parts:
            yield part

    def __len__(self):
        return len(self.parts)

    def __eq__(self, other):
        if not isinstance(other, DottedName):
            return False
        return self.parts == other.parts

    def __objpath__(self):
        return '.'.join(self.parts)

    def __str__(self):
        return self.__objpath__()

    def __repr__(self):
        return f"{DottedName.__name__}('{self}')"

    def __hash__(self):
        return hash(objpath(self))

    def __dhash__(self):
        return dhash(objpath(self))

    @staticmethod
    def _split_path(path):
        if isinstance(path, DottedName):
            return path.parts.copy()
        else:
            if path == '':
                return []
            else:
                return objpath(path).split('.')


class ObjectPath:
    def __init__(self, path, nested_path=None):
        """用于表示一个Python对象路径

        支持的形式包括：
            * '.'分割路径，例如xxx.yyy.zzz.Target
            * '.'分割路径但用':'分割模块和模块内路径，例如xxx.yyy.zzz:Target

        使用resolve后

        Parameters
        ----------
        path
            模块名或目标完整路径
        nested_path
            目标的模块内路径，如果未定义则认为path为完整名，会自动识别path的路径形式
        """
        self.module = None
        if isinstance(path, module_type):
            mod = path
            self.module = mod
            self.set_module_name_by_module(mod)
            self.nested_path = nested_path
        else:
            if nested_path is None:
                self.module_name, self.nested_path = split_colon_seperated_path(path)
            else:
                self.module_name = path
                self.nested_path = nested_path

    @staticmethod
    def of(obj_or_path):
        if isinstance(obj_or_path, str):
            ret = ObjectPath(obj_or_path)
        else:
            module = inspect.getmodule(obj_or_path)
            if module is None:
                raise ValueError(f'Unable to locate which module contains "{obj_or_path}".')
            nested_path = getattr(obj_or_path, '__qualname__', None)
            if nested_path is None:
                raise ValueError(f'"{obj_or_path}" does not contain "__qualname__". '
                                 f'It looks like an instance instead of valid targets like type or function.')
            ret = ObjectPath(path=module, nested_path=nested_path)
        return ret

    @property
    def parent(self):
        if len(self.nested_path) > 0:
            mod = self.module if self.module is not None else self.module_name
            if len(self.nested_path) > 1:
                return ObjectPath(mod, self.nested_path.parent)
            else:
                return ObjectPath(mod)
        else:
            return ObjectPath(self.module_name.parent)

    def resolve(self):
        ret = None

        if self.is_colon_seperated:
            if not self.is_module_loaded:
                # 此时nest path是正常状态，不需要重新修改
                self.module = importlib.import_module(objpath(self.module_name))
        else:
            ret = resolve_name_compat(objpath(self.nested_path))
            if isinstance(ret, module_type):
                module = ret
            else:
                import inspect
                module = inspect.getmodule(ret)

            self.module = module
            self.set_module_name_by_module(module)
            self.trim_nest_path_by_module_name()

        if ret is None and self.is_module_loaded:
            if self.is_module:
                ret = self.module
            else:
                ret = get_object_from(self.module, objpath(self.nested_path))

        return ret

    @property
    def module(self):
        return self._module

    @module.setter
    def module(self, mod):
        self._module = mod

    @property
    def module_name(self):
        return self._module_name

    @module_name.setter
    def module_name(self, val):
        if val is None:
            val = ''
        self._module_name = DottedName(val)

    @property
    def nested_path(self):
        return self._nested_path

    @nested_path.setter
    def nested_path(self, val):
        if val is None:
            val = ''
        self._nested_path = DottedName(val)

    def set_module_name_by_module(self, mod):
        self.module_name = mod.__name__

    def trim_nest_path_by_module_name(self):
        self.nested_path = self.nested_path.without_prefix(self.module_name)

    @property
    def is_colon_seperated(self):
        return not self.module_name.empty

    @property
    def is_module(self) -> Optional[bool]:
        if self.is_colon_seperated:
            return self.nested_path.empty
        else:
            return None

    @property
    def is_module_loaded(self):
        if self.module is not None:
            return True

        if self.is_colon_seperated:
            module = sys.modules.get(objpath(self.module_name), None)
            if module is not None:
                self.module = module
                self.set_module_name_by_module(module)
                self.trim_nest_path_by_module_name()
                return True

        return False

    def __iter__(self):
        for part in self.module_name:
            yield part
        for part in self.nested_path:
            yield part

    def __eq__(self, other):
        if not isinstance(other, ObjectPath):
            return False

        if self.is_module_loaded and other.is_module_loaded:
            return self.module == other.module and self.nested_path == other.nested_path

        for lhs, rhs in zip(self, other):
            if lhs != rhs:
                return False
        return True

    def __objpath__(self):
        if self.is_colon_seperated:
            return ':'.join((objpath(self.module_name), objpath(self.nested_path)))
        else:
            return objpath(self.nested_path)

    def __str__(self):
        return self.__objpath__()

    def __repr__(self):
        if self.is_colon_seperated:
            return f"{ObjectPath.__name__}(path='{self.module_name}', nested_path='{self.nested_path}')"
        else:
            return f"{ObjectPath.__name__}(path='{self.nested_path}')"

    def to_dotted(self):
        if self.is_colon_seperated:
            return DottedName(self.module_name, self.nested_path)
        else:
            return DottedName(self.nested_path)

    def __hash__(self):
        return hash(self.to_dotted())

    def __dhash__(self):
        return dhash(self.to_dotted())


objpath_like = Union[DottedName, ObjectPath, str]


def objpath(obj) -> str:
    """获取obj所代表的Python对象路径，类似os.fspath

    Parameters
    ----------
    obj
        代表Python对象路径的对象

    Returns
    -------
    ret
        obj所代表的对象路径

    Notes
    -----
        如果obj实现了__objpath__，则会调用该方法获取返回结果
        否则会使用str尝试将其转换为字符串，并触发警告

    Warnings
    --------
        objpath并非获取obj的对象路径
    """
    ret = getattr(obj, DottedName.__objpath__.__name__, None)
    if ret is not None:
        ret = ret()
    elif isinstance(obj, str):
        ret = obj
    else:
        warnings.warn(f'Object `{get_type_name(obj)}` is not convertible to dotted name. Will try using str(), '
                      f'which may lead to unexpected result.')
        ret = str(obj)

    return ret


def split_colon_seperated_path(path: str):
    """用于分割使用':'定义的Python对象路径，':'分割了模块和模块内路径

    Parameters
    ----------
    path
        路径

    Returns
    -------
    ret: (str or None, str)
        如果对象路径包含冒号':'，则从冒号处分割并返回 (str_before_colon, str_after_colon)，否则返回 (None, path)
    """
    segs = path.split(':', maxsplit=1)
    if len(segs) == 2:
        return tuple(segs)
    else:
        return None, path


def get_object_from(source, path: str):
    """用于进行从module或者其他对象中提取类似 'xxx.yyy.zzz' 路径的对象

    Parameters
    ----------
    source
        源，比如module或class
    path
        待获取的对象路径

    Returns
    -------
    ret
        获取到的对象，或None
    """
    ret = source

    if path == '':
        return ret

    for part in path.split('.'):
        ret = getattr(ret, part, None)
        if ret is None:
            break

    return ret


def get_qualname(obj):
    """获取目标的模块内路径，如果不存在__qualname__属性则返回__name__属性

    Parameters
    ----------
    obj
        目标对象

    Returns
    -------
    ret
        目标的模块内路径
    """
    ret = getattr(obj, '__qualname__', None)
    if ret is None:
        ret = getattr(obj, '__name__', None)

    return ret


def get_module_name(obj):
    """获取目标关联的模块名，一般是__module__

    Parameters
    ----------
    obj
        目标对象

    Returns
    -------
    ret
        目标关联的模块名
    """
    return getattr(obj, '__module__', None)


def get_full_qualified_path(obj):
    """从目标对象本身获取目标的包括模块名在内的全限定路径，使用 : 分割包和包内路径

    Parameters
    ----------
    obj
        目标对象，包括类或函数

    Returns
    -------
    ret
        目标对象的全限定路径字符串

    Examples
    --------
    >>> get_full_qualified_path(xxx.yyy.zzz.Target)
    <<< 'xxx.yyy.zzz:Target'

    Warnings
    --------
        该函数只接受类或函数等具有__module__和__qualname__的目标，常规类的实例无法以此方法获得路径
    """
    return objpath(ObjectPath.of(obj))


def reassign_object_name(obj, new_name: objpath_like = None, new_qualname: objpath_like = None):
    """重新给obj赋予名字，同时修改__name__和__qualname__，如果不存在则添加

    Parameters
    ----------
    obj
        待改名的对象
    new_name
        新的名字
    new_qualname
        新的模块内全限定名
    """
    if new_name is not None:
        new_name = DottedName(new_name)
        obj.__name__ = objpath(new_name)
        qualname = getattr(obj, '__qualname__', None)
        if qualname is not None:
            qualname = DottedName(qualname)
            qualname[-1] = new_name
            obj.__qualname__ = objpath(qualname)
    elif new_qualname is not None:
        new_qualname = DottedName(new_qualname)
        obj.__qualname__ = objpath(new_qualname)
        obj.__name__ = objpath(new_qualname.suffix(1))
    else:
        warnings.warn(f'{reassign_object_name.__name__} called without giving new names.')


def reassign_object_module(obj, new_module: objpath_like):
    """重新给obj赋予所关联的模块，如果不存在__module__则会直接添加

    Parameters
    ----------
    obj
        待修改关联模块的对象
    new_module
        新的所属模块名
    """
    obj.__module__ = objpath(new_module)


def get_nest_prefix_by_qualname(obj) -> str:
    """通过解析__qualname__提取对象在模块中的嵌套路径，如果是模块下直接定义的对象，则返回空字符串

    Parameters
    ----------
    obj
        待获取嵌套路径的对象，可以为type、类实例或方法以及其他具有__qualname__的对象

    Returns
    -------
    ret
        如果目标属于某嵌套定义下，则返回__qualname__中提取的嵌套路径（不包含对象本身的名字），
        如果是模块下直接定义的对象，则返回空字符串
    """
    segments = obj.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)
    return '.'.join(segments[:-1])


def get_object_by_path(path):
    """通过全限定路径获取目标

    Parameters
    ----------
    path
        目标的全限定路径

    Returns
    -------
    ret
        全限定路径所标识的目标

    Examples
    --------
    >>> get_object_by_path('xxx.yyy.zzz.Target')
    <class 'xxx.yyy.zzz.Target'>
    """
    return ObjectPath(path).resolve()


def get_class_that_defined_method(meth):
    """获取定义指定方法的类

    Parameters
    ----------
    meth
        方法实例

    Returns
    -------
    ret
        定义meth的类，或者None，代表该方法无法找到

    Notes
    -----
        由于Python3移除了所有类和类中定义的方法的直接关联，因此只能通过一些方法来“猜测”，所以理论上存在无法获取到正确结果的情况

        比如如果随意修改__qualname__或者__module__的值会导致结果不可预测

    References
    ----------
        By @Yoel and with assistance from comments

        "Get class from meth.__globals__" patch by @Alexander_McFarlane

        https://stackoverflow.com/questions/3589311/get-defining-class-of-unbound-method-object-in-python-3/

        Modified to support statically nested class using 'get_object_from'
    """
    import functools
    if isinstance(meth, functools.partial):
        return get_class_that_defined_method(meth.func)
    if inspect.ismethod(meth) or (inspect.isbuiltin(meth) and
                                  getattr(meth, '__self__', None) is not None and
                                  getattr(meth.__self__, '__class__', None)):
        for cls in inspect.getmro(meth.__self__.__class__):
            if meth.__name__ in cls.__dict__:
                return cls
        meth = getattr(meth, '__func__', meth)  # fallback to __qualname__ parsing
    if inspect.isfunction(meth):
        class_prefix = get_nest_prefix_by_qualname(meth)
        cls = get_object_from(inspect.getmodule(meth), class_prefix)
        if cls is None:
            path_segs = class_prefix.split('.')
            cls = get_object_from(meth.__globals__.get(path_segs[0]), '.'.join(path_segs))
        if isinstance(cls, type):
            return cls
    return getattr(meth, '__objclass__', None)  # handle special descriptor objects


def get_nest(obj):
    """获取对象所在的类或模块

    Parameters
    ----------
    obj
        对象

    Returns
    -------
    ret
        方法所在的类

    Notes
    -----
        由于Python3移除了所有类和类中定义的方法以及和模块的直接关联，因此只能通过一些方法来“猜测”，所以理论上存在无法获取到正确结果的情况

        比如如果随意修改__qualname__或者__module__的值会导致结果不可预测
    """
    if inspect.ismethod(obj):
        return obj.__self__
    elif inspect.isclass(obj):
        module = inspect.getmodule(obj)
        path = get_nest_prefix_by_qualname(obj)
        return get_object_from(module, path)
    else:
        ret = get_class_that_defined_method(obj)
        if ret is None:
            if obj.__qualname__ == obj.__name__:
                # 猜测定义在模块中
                ret = inspect.getmodule(obj)

        return ret


def get_type_name(instance):
    return type(instance).__name__


_NAME_PATTERN = None


def _resolve_name_compat(name):
    """解析path定义的对象路径，并返回路径

    实现来自Python 3.9的pkgutil.resolve_name

    Parameters
    ----------
    name
        Python对象路径

    Returns
    -------
    ret
        目标对象
    """
    global _NAME_PATTERN
    if _NAME_PATTERN is None:
        import re
        _DOTTED_WORDS = r'(?!\d)(\w+)(\.(?!\d)(\w+))*'
        _NAME_PATTERN = re.compile(f'^(?P<pkg>{_DOTTED_WORDS})(?P<cln>:(?P<obj>{_DOTTED_WORDS})?)?$', re.U)

    m = _NAME_PATTERN.match(name)
    if not m:
        raise ValueError(f'invalid format: {name!r}')
    gd = m.groupdict()
    if gd.get('cln'):
        # there is a colon - a one-step import is all that's needed
        mod = importlib.import_module(gd['pkg'])
        parts = gd.get('obj')
        parts = parts.split('.') if parts else []
    else:
        # no colon - have to iterate to find the package boundary
        parts = name.split('.')
        modname = parts.pop(0)
        # first part *must* be a module/package.
        mod = importlib.import_module(modname)
        while parts:
            p = parts[0]
            s = f'{modname}.{p}'
            try:
                mod = importlib.import_module(s)
                parts.pop(0)
                modname = s
            except ImportError:
                break
    # if we reach this point, mod is the module, already imported, and
    # parts is the list of parts in the object hierarchy to be traversed, or
    # an empty list if just the module is wanted.
    result = mod
    for p in parts:
        result = getattr(result, p)
    return result


if sys.version_info >= (3, 9):
    from pkgutil import resolve_name
    resolve_name_compat = resolve_name
else:
    resolve_name_compat = _resolve_name_compat



from typing import Union

import numpy as np
import pyvista as pv

from metalpy.utils.model import extract_model_list_bounds


class SubplotSwitched:
    def __init__(self, plotter: pv.BasePlotter, index_loc=None):
        self.plotter = plotter
        self.backup = plotter.renderers.index_to_loc(plotter.renderers.active_index)
        if index_loc is not None and isinstance(index_loc, int):
            index_loc = [index_loc, None]

        if np.all(index_loc == self.backup):
            self.target = None
        else:
            self.target = index_loc

    def __enter__(self):
        if self.target is not None:
            self.plotter.subplot(*self.target)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.target is not None:
            self.plotter.subplot(*self.backup)


class SliceView:
    def __init__(self, plotter: pv.BasePlotter, index_loc=None):
        """实现一个三维模型的切片视图

        Notes
        -----
        初始化后使用绑定主视角鼠标事件
            plotter.enable_cell_picking(callback=slice_view.clip_by_selection, show=False, through=False)

        操作方式为：

        `p`: 切分所点击网格所在正交平面
        `r`: 切分选定区域内的网格

        Parameters
        ----------
        plotter
            绑定的绘制设备

        index_loc
            绑定的视图坐标

        Examples
        --------
        >>> plotter = pv.Plotter(shape=(1, 2))
        >>> plotter.subplot(0, 0)
        >>> slice_view = SliceView(plotter, [0, 1])

        >>> voxels = pv.voxelize(pv.Sphere(1, (0, 0, 0)).scale((8, 4, 2), inplace=False), density=0.4)
        >>> plotter.add_mesh(voxels, opacity=1, style='surface', show_edges=True)
        >>> slice_view.add_mesh(voxels, opacity=1, style='surface', show_edges=True)

        >>> plotter.show_grid()
        >>> plotter.enable_cell_picking(callback=slice_view.clip_by_selection, show=False, through=False)

        >>> plotter.show()
        """
        self.mesh_elements = []
        self.actors = []
        self.plotter = plotter
        self.index_loc = index_loc

    def switched(self):
        """将绘制设备的活动视图临时改为绑定的子视图

        Returns
        -------
        ret : SubplotSwitched
            用于with块
        """
        return SubplotSwitched(self.plotter, self.index_loc)

    def _add_mesh(self, mesh, **kwargs):
        ret = self.plotter.add_mesh(mesh, **kwargs)
        self.actors.append(ret)

    def add_mesh(self, mesh, **kwargs):
        """同Plotter.add_mesh

        Parameters
        ----------
        mesh
            待切片的mesh

        kwargs
            其它传给Plotter.add_mesh的参数
        """
        self.mesh_elements.append((mesh, kwargs))

        with self.switched():
            self._add_mesh(mesh, **kwargs)

    def clear(self):
        """清理切片视图中绘制的切片
        """
        with self.switched():
            for t in self.actors:
                self.plotter.remove_actor(t)
            self.actors.clear()

    def clip(self, bounds, refresh=True, update=True):
        """使用指定的边界截取切片

        Parameters
        ----------
        bounds
            切片边界

        refresh
            是否清理已有的切片图形

        update
            是否自动更新视图（如果有其它需要更新的内容可以置为False然后在函数外手动更新）
        """
        with self.switched():
            if refresh:
                self.clear()

            for mesh, kwargs in self.mesh_elements:
                if bounds is not None:
                    mesh = mesh.clip_box(bounds, invert=False)
                if mesh.n_cells > 0:
                    self._add_mesh(mesh, **kwargs)

            if update:
                self.plotter.update(force_redraw=True)

    def clip_by_selection(self, selection: Union[pv.UnstructuredGrid, pv.MultiBlock],
                          enable_ortho_slice=True,
                          refresh=True, update=True):
        """使用选定区域与视角来判断截取切片的范围用于切片

        界面确定方法为:
        1. 视角所朝向的轴为延申轴a
        2. 截取范围中相对场景范围较宽的方向为延申轴b
        3. 截取限定第三轴范围和selection一致下，a-b平面上的切片

        若启用正交切片enable_ortho_slice
        则在第一步后，分别以除a外的两轴进行一次切片，组合展示

        Parameters
        ----------
        selection
            切片边界

        enable_ortho_slice
            是否启用p键单cell选择时展示正交切片

        refresh
            是否清理已有的切片图形

        update
            是否自动更新视图（如果有其它需要更新的内容可以置为False然后在函数外手动更新）
        """
        if refresh:
            self.clear()

        if selection is None:
            print('Bad Selection!')
            return

        bounds = np.asarray(selection.bounds)
        camera_pos = np.asarray(self.plotter.camera_position[0])
        axis = np.argmax(abs(camera_pos))

        scene_bounds = np.asarray(extract_model_list_bounds([a[0] for a in self.mesh_elements]))
        scene_sizes = scene_bounds[1::2] - scene_bounds[::2]
        sizes = bounds[1::2] - bounds[::2]
        sizes[axis] = 0
        axis2 = np.argmax(sizes / scene_sizes)
        print(axis, axis2)

        if axis == axis2:
            print('Bad Selection!')
            return

        bounded_axis = 0 + 1 + 2 - axis - axis2
        clipping_bounds = scene_bounds.copy()
        clipping_bounds[2 * bounded_axis:2 * bounded_axis + 2] = \
            bounds[2 * bounded_axis:2 * bounded_axis + 2]

        self.clip(clipping_bounds, refresh=False, update=False)
        if enable_ortho_slice and \
                isinstance(selection, pv.UnstructuredGrid) and \
                selection.n_cells == 1:
            clipping_bounds = scene_bounds.copy()
            clipping_bounds[2 * axis2:2 * axis2 + 2] = \
                bounds[2 * axis2:2 * axis2 + 2]
            self.clip(clipping_bounds, refresh=False, update=False)

            prev_pos = self.plotter.camera_position
            with self.switched():
                self.plotter.camera_position = prev_pos
        else:
            with self.switched():
                self.plotter.camera_position = ['yz', 'xz', 'xy'][bounded_axis]

        if update:
            self.plotter.update(force_redraw=True)



from __future__ import annotations

import os
import traceback
import warnings
from functools import cache
from pathlib import Path
from typing import Union, TYPE_CHECKING

from .object_path import ObjectPath, objpath
from .type import ensure_as_iterable, undefined

if TYPE_CHECKING:
    from datetime import timedelta, datetime


PathLikeType = (str, os.PathLike)
PathLike = Union[str, os.PathLike]
cache_dir = Path('./.cache')


def ensure_dir(path: PathLike):
    path = Path(path)
    if not path.exists():
        path.mkdir(parents=True, exist_ok=True)
    return os.fspath(path)


def ensure_filepath(path: PathLike):
    path = Path(path)
    ensure_dir(path.parent)
    return os.fspath(path)


def locate_file(path, *predicates, mode='first'):
    """快速查找特定的文件
    :param path: 查找路径
    :param predicates: 判断谓词
    :param mode: 返回模式
        all - 返回全部符合条件的文件
        first - 返回第一个符合条件的文件
        last - 返回最后一个符合条件的文件
    :return: mode为all时，返回一个符合条件的完整文件路径列表（可能为空），否则返回一个完整文件路径或None
    """
    immediate_return = True
    ret = []
    if mode == 'all':
        immediate_return = False

    iterable = os.listdir(path)
    if mode == 'last':
        iterable = reversed(iterable)

    for file in iterable:
        flag = True
        for predicate in predicates:
            flag = flag and predicate(file)
            if not flag:
                break

        if flag:
            ret.append(file)
            if immediate_return:
                break

    ret = [os.path.join(path, file) for file in ret]
    if immediate_return:
        return ret[0] if len(ret) > 0 else None
    else:
        return ret


def locate_prefixed_file(path, prefix):
    return locate_file(path, lambda file: file.startswith(prefix))


def locate_file_by(paths, *, prefix=None, suffix=None, ext=None, contains=None, filename=None, mode='all'):
    preds = []

    if prefix is not None:
        prefix = ensure_as_iterable(prefix, str)
        preds.append(lambda f: any(map(lambda s: f.startswith(s), prefix)))

    if suffix is not None:
        suffix = ensure_as_iterable(suffix, str)
        preds.append(lambda f: any(map(lambda s: f.endswith(s), suffix)))

    if ext is not None:
        ext = ensure_as_iterable(ext, str)
        preds.append(lambda f: os.path.splitext(f)[1] in ext)

    if contains is not None:
        contains = ensure_as_iterable(contains, str)
        preds.append(lambda f: any(map(lambda s: s in f, contains)))

    if filename is not None:
        filename = ensure_as_iterable(filename, str)
        preds.append(lambda f: f in filename)

    ret = []

    immediate_return = True
    if mode == 'all':
        immediate_return = False

    paths = ensure_as_iterable(paths, str)
    filepaths = None
    for path in paths:
        filepaths = locate_file(path, *preds, mode=mode)
        if immediate_return:
            if filepaths is not None:
                break
        else:
            ret.extend(filepaths)

    if immediate_return:
        return filepaths
    else:
        return ret


def locate_first_file_by(paths, *, prefix=None, suffix=None, ext=None, contains=None, filename=None):
    return locate_file_by(paths, prefix=prefix, suffix=suffix, ext=ext, contains=contains, filename=filename,
                          mode='first')


def locate_last_file_by(paths, *, prefix=None, suffix=None, ext=None, contains=None, filename=None):
    return locate_file_by(paths, prefix=prefix, suffix=suffix, ext=ext, contains=contains, filename=filename,
                          mode='last')


def locate_files_by(paths, *, prefix=None, suffix=None, ext=None, contains=None, filename=None):
    return locate_file_by(paths, prefix=prefix, suffix=suffix, ext=ext, contains=contains, filename=filename,
                          mode='all')


def git_ignore_directory(path):
    ignore_file = (Path(path) / '.gitignore')
    if not ignore_file.exists():
        ignore_file.write_text('# Created by metalpy automatically.\n*')


def make_cache_file(name, *args):
    return os.fspath(make_cache_file_path(name, *args))


def make_cache_directory(name, *args):
    return os.fspath(make_cache_directory_path(name, *args))


def make_cache_file_path(name, *args):
    ret = Path(cache_dir, name, *args).absolute()
    ensure_filepath(ret)
    git_ignore_directory(cache_dir)
    return ret


def make_cache_directory_path(name, *args):
    ret = Path(cache_dir, name, *args).absolute()
    ensure_dir(ret)
    git_ignore_directory(cache_dir)
    return ret


class FileCache:
    def __init__(self, func, ttl: 'timedelta' = None):
        self.func = func
        self.ttl = ttl

    def __call__(self, *args, **kwargs):
        return self.prepare()(*args, **kwargs)

    def force_update(self):
        return self.prepare(update=True)

    def prepare(self, update=False):
        return FileCachedCall(self, update=update)


class FileCachedCall:
    def __init__(self, cache_obj: FileCache, update=False):
        self.cache = cache_obj
        self.update = update

    @property
    def func(self):
        return self.cache.func

    @property
    def ttl(self):
        return self.cache.ttl

    @cache
    def get_func_namespace(self):
        return objpath(ObjectPath.of(self.func).nested_path)

    def __call__(self, *args, **kwargs):
        from metalpy.utils.dhash import dhash
        name = self.get_func_namespace()
        hashkey = dhash(self.func, *args, kwargs)

        val = undefined
        if not self.update:
            val = get_cache(hashkey, ttl=self.ttl, name=name)

        if val is not undefined:
            return val
        else:
            val = self.func(*args, **kwargs)
            put_cache(hashkey, val, name=name)
            return val


def get_cache_key(key, name: str | None = None):
    from metalpy.utils.dhash import dhash
    code = dhash(key).hexdigest(32)
    if name is not None:
        return f'{name}[{code}]'
    else:
        return code


def get_cache_path(key, name: str | None = None):
    return make_cache_directory_path('cached') / get_cache_key(key, name=name)


def put_cache(key, content, name: str | None = None):
    import cloudpickle

    file = get_cache_path(key, name=name)
    ensure_filepath(file)
    with file.open('wb') as f:
        cloudpickle.dump((key, content), f)


def get_cache(key, default=undefined, ttl: 'timedelta' = None, name: str | None = None):
    from datetime import datetime
    import cloudpickle

    file = get_cache_path(key, name=name)

    if file.exists():
        if ttl is not None:
            if datetime.now() - datetime.fromtimestamp(os.path.getmtime(file)) > ttl:
                return undefined

        with file.open('rb') as f:
            try:
                cache_key, content = cloudpickle.load(f)
                if cache_key == key:
                    return content
            except Exception:
                warnings.warn(traceback.format_exc() +
                              '\nException occurred when loading cache. Ignoring cached file.')

    return default


def clear_cache(key, name: str | None = None):
    file = get_cache_path(key, name=name)
    file.unlink(missing_ok=True)


def file_cached(func=None, ttl: 'timedelta' = None):
    import functools
    if func is None:
        return functools.partial(file_cached, ttl=ttl)
    else:
        return functools.wraps(func)(FileCache(func, ttl=ttl))


def openable(path):
    return isinstance(path, PathLikeType)



import numpy as np
import psutil
import taichi as ti

from metalpy.scab.utils.misc import Field
from metalpy.utils.ti_solvers import matrix_free
from metalpy.utils.taichi import ti_kernel, ti_field, ti_FieldsBuilder, ti_func, copy_from
from .integrated import solve_Ax_b
from .kernel import kernel_matrix_forward
from .solver import DemagnetizationSolver


class SeperatedSolver(DemagnetizationSolver):
    def __init__(
            self,
            receiver_locations: np.ndarray,
            xn: np.ndarray,
            yn: np.ndarray,
            zn: np.ndarray,
            base_cell_sizes: np.ndarray,
            source_field: Field,
            direct_to_host: bool,
            is_cpu: bool,
            progress: bool
    ):
        """该函数将矩阵分为9个部分，从而实现一些优化

        Parameters
        ----------
        receiver_locations
            观测点
        xn, yn, zn
            网格边界
        base_cell_sizes
            网格最小单元大小
        direct_to_host
            是否通过kernel将结果直接zero-copy地复制到待返回的numpy数组
        is_cpu
            是否是CPU架构，如果是，才会在内存可用时合并核矩阵然后求解
        progress
            是否输出求解进度条，默认为False不输出

        Notes
        -----
        优势：

        - 可以一定程度上绕过完整核矩阵尺寸的int32限制 (允许的网格规模为 `Integrated` 方法的3倍)
        - 可以使用AoS提高内存近邻性

        缺陷：

        - 网格规模仍然受到taichi的int32索引限制
        - 相比于 `Integrated` 方法，内存需求并不会降低，并且如果矩阵规模较小，计算效率可能会低于 `Integrated` 方法
        """
        super().__init__(receiver_locations, xn, yn, zn, base_cell_sizes, source_field)
        self.direct_to_host = direct_to_host
        self.is_cpu = is_cpu
        self.progress = progress

        self.builder = builder = ti_FieldsBuilder()

        nC = xn.shape[0]
        nObs = receiver_locations.shape[0]

        self.Tmat321 = [
            [ti_field(self.kernel_type) for _ in range(3)],  # Txx, Txy, Txz
            [ti_field(self.kernel_type) for _ in range(2)],  # ---- Tyy, Tyz
            [ti_field(self.kernel_type) for _ in range(1)],  # --------- Tzz
        ]
        self.Tmat33 = [
            self.Tmat321[0],  # Txx, Txy, Txz
            [self.Tmat321[0][1]] + self.Tmat321[1],  # Tyx, Tyy, Tyz
            [self.Tmat321[0][2], self.Tmat321[1][1]] + self.Tmat321[2]  # Tzx, Tzy, Tzz
        ]
        self.Tmat6 = [t for ts in self.Tmat321 for t in ts]

        builder.dense(ti.ij, (nObs, nC)).place(*self.Tmat6)

        builder.finalize()

    def build_kernel(self, model):
        kernel_matrix_forward(
            self.receiver_locations,
            self.xn, self.yn, self.zn,
            self.base_cell_sizes, model,
            *self.Tmat6, np.empty(0),
            write_to_mat=False, compressed=False
        )

    def solve(self, magnetization):
        if self.is_cpu and psutil.virtual_memory().percent < 45:
            # TODO: 进一步考虑模型大小来选择求解方案，模型较大时也应该使用 `solve_Tx_b`
            return solve_Ax_b(merge_Tmat_as_A(self.Tmat33, direct_to_host=self.direct_to_host), magnetization)
        else:
            return solve_Tx_b(self.Tmat321, magnetization, progress=self.progress)


def solve_Tx_b(Tmat321, m, progress: bool = False):
    with ti_FieldsBuilder() as builder:
        x = builder.place_dense_like(m[:, None])
        b = builder.place_dense_like(m[:, None])

        builder.finalize()

        copy_from(b, m[:, None])

        dtype = x.dtype

        @ti_func
        def extract(r, c):
            return (
                Tmat321[0][0][r, c], Tmat321[0][1][r, c], Tmat321[0][2][r, c],
                Tmat321[1][0][r, c], Tmat321[1][1][r, c],
                Tmat321[2][0][r, c]
            )

        @ti_kernel
        def linear(x: ti.template(), Ax: ti.template()):
            n_obs = Ax.shape[0] // 3
            n_cells = x.shape[0] // 3

            for r in range(n_obs):
                row = r * 3

                bx: dtype = 0
                by: dtype = 0
                bz: dtype = 0

                for c in range(0, n_cells):
                    col = c * 3

                    txx, txy, txz, tyy, tyz, tzz = extract(r, c)
                    mx, my, mz = x[col + 0, 0], x[col + 1, 0], x[col + 2, 0]

                    bx += txx * mx + txy * my + txz * mz
                    by += txy * mx + tyy * my + tyz * mz
                    bz += txz * mx + tyz * my + tzz * mz

                Ax[row + 0, 0] = bx
                Ax[row + 1, 0] = by
                Ax[row + 2, 0] = bz

        matrix_free.cg(ti.linalg.LinearOperator(linear), b, x, progress=progress)

        return x.to_numpy()


def merge_Tmat_as_A(Tmat33, direct_to_host=True):
    nObs, nC = Tmat33[0][0].shape
    Amat = np.empty((3 * nObs, 3 * nC), dtype=np.float64)
    for i in range(3):
        for j in range(3):
            if direct_to_host:
                tensor_to_ext_arr(Tmat33[i][j], Amat, i, 3, j, 3)
            else:
                Amat[i::3, j::3] = Tmat33[i][j].to_numpy()

    return Amat


@ti_kernel
def tensor_to_ext_arr(tensor: ti.template(), arr: ti.types.ndarray(),
                      x0: ti.template(), xstride: ti.template(),
                      y0: ti.template(), ystride: ti.template()):
    for I in ti.grouped(tensor):
        arr[I[0] * xstride + x0, I[1] * ystride + y0] = tensor[I]


@ti_func
def multiply_with_row_stride_3(
        out: ti.template(),
        A: ti.template(),
        B: ti.template(),
        C: ti.template(),
        v: ti.template(),
        row0: ti.template(), row_stride: ti.template()
):
    for r in range(A.shape[0]):
        row = r * row_stride + row0

        summed = (
                A[r, 0] * v[0, 0]
                + B[r, 0] * v[1, 0]
                + C[r, 0] * v[2, 0]
        )

        for c in range(1, A.shape[1]):
            col = c * 3
            summed += (
                    A[r, c] * v[col + 0, 0]
                    + B[r, c] * v[col + 1, 0]
                    + C[r, c] * v[col + 2, 0]
            )

        out[row, 0] = summed



import warnings
from functools import cached_property
from itertools import product
from typing import Iterable

import numpy as np

from metalpy.utils.bounds import Bounds
from . import Shape3D, Cuboid, Composition


def clarify_spec(spec, n_params=3):
    """定义一个长度为n_params的数组

    如果spec为可迭代类型，则验证spec长度是否满足。

    如果spec不为可迭代类型，则认为用单个spec定义了数组中的所有值，返回[spec] * n_params

    Parameters
    ----------
    spec
        待验证的参数
    n_params
        期望的参数个数

    Returns
    -------
    ret
        长度为n_params的数组，其值由spec定义
    """
    if isinstance(spec, Iterable):
        spec = list(spec)
        assert len(spec) == n_params
    else:
        spec = [spec] * n_params

    return np.asarray(spec)


class BarFramework(Composition):
    def __init__(self,
                 outline: Shape3D,
                 spec=None,
                 n_rooms=None,
                 room_spec=None,
                 bounds=None,
                 *,
                 inherit_transform=False,
                 bar_spec=None
                 ):
        """定义一个规则框架模型，其边界由 `outline` 决定，并基于 `outline` 裁剪边界。

        Parameters
        ----------
        outline
            边界模型，一般为Cuboid或者立方体Prism
        spec, bar_spec
            三个方向上条柱的粗细，传入单一值时代表同时定义三个方向
        n_rooms
            三个方向上的房间数，传入单一值时代表同时定义三个方向
        room_spec
            三个方向上的房间尺寸，传入单一值时代表同时定义三个方向。
            房间尺寸以条柱中心为基准计算，保证结果中的房间尺寸为最大的不大于room_spec的等距划分。
            例如x方向有效宽度为12m，约定房间尺寸为11m，则结果中x方向房间数为2，单个房间尺寸为6m。
        bounds
            覆盖边界模型的边界。
            如果 `inherit_transform=True` 则应提供 `outline` 本地坐标系下的边界坐标。
            否则应提供世界坐标系下的边界。
        inherit_transform
            指示框架模型是否从 `outline` 继承空间变换信息。
            为 `True` 时框架模型边界将随着 `outline` 进行变换（会因变换而不和坐标轴正交）。
            为 `False` 时框架模型边界不会随着 `outline` 进行变换（保持与坐标轴正交）。
        """
        if spec is not None:
            self.spec = spec
        elif bar_spec is not None:
            self.bar_spec = bar_spec
        else:
            raise AssertionError('`spec` must be specified.')

        if bounds is not None:
            bounds = Bounds(bounds)
            if inherit_transform:
                # 通过 `Cuboid` 做坐标转换
                bounds = Cuboid(bounds=bounds).apply(outline.transforms, inplace=True).bounds

            self.outline_bounds = bounds

        if n_rooms is None:
            # 假设使用房间的尺寸定义
            # 则保证生成的房间不大于指定的尺寸
            # ni = ceil(bi / ri)
            room_spec = clarify_spec(room_spec)
            # 约定条柱不会超出outline定义的边界
            # 因此从边界条柱的中心开始算，实际的有效空间大小有缩小
            sizes = outline.local_bounds.extent - self.spec
            self.n_rooms = np.ceil(sizes / room_spec).astype(int)
        else:
            self.n_rooms = n_rooms

        self.outline = outline
        self.inherit_transform = inherit_transform

        # Shape3D层面保证所有place的结果为正值，所以Min和Max可以分别作为交集和并集使用
        super().__init__(
            mix_mode=Composition.Intersects
        )

    @property
    def spec(self):
        return self._spec

    @spec.setter
    def spec(self, v):
        self._spec = clarify_spec(v, 3)

    @property
    def n_rooms(self):
        return self._n_rooms

    @n_rooms.setter
    def n_rooms(self, v):
        self._n_rooms = clarify_spec(v, 3)

    @property
    def n_bars(self):
        return self.n_rooms + 1

    @property
    def bar_spec(self):
        return self.spec

    @bar_spec.setter
    def bar_spec(self, v):
        warnings.warn('`bar_spec` has been deprecated, consider using `spec` intead.')
        self.spec = v

    @property
    def bar_radius(self):
        return self.bar_spec / 2

    @property
    def axis_span(self):
        return self.outline_bounds.view().reshape((-1, 2))

    def with_(self, spec=None, n_rooms=None, outer_bars_only=False, *, bar_spec=None):
        if spec is not None:
            self.spec = spec
        elif bar_spec is not None:
            self.bar_spec = bar_spec

        if n_rooms is not None:
            self.n_rooms = n_rooms
        elif outer_bars_only:
            self.n_rooms = 1

        return self

    def __getitem__(self, indices):
        full = slice(None, None, None)

        n_rooms = self.n_rooms
        bounds = self.local_bounds

        for axis, slicer in enumerate(indices):
            if slicer == full:
                continue

            if np.isscalar(slicer):
                slicer = slice(slicer, slicer + 1)

            centers = self.centers[axis]
            r = self.bar_radius[axis]

            lower_bars = centers[:-1][slicer]
            upper_bars = centers[1:][slicer]
            bounds.set(
                axis,
                min=min(lower_bars) - r,
                max=max(upper_bars) + r,
            )

            n_rooms[axis] = len(lower_bars)

        return BarFramework(outline=self.outline, bar_spec=self.bar_spec, n_rooms=n_rooms, bounds=bounds)

    def get_centers(self, axis):
        a0, a1 = self.axis_span[axis]
        ra = self.bar_radius[axis]
        na = self.n_bars[axis]
        return np.linspace(a0 + ra, a1 - ra, na)

    @cached_property
    def outline_bounds(self):
        """获取框架部分在局部坐标系下的边界，框架部分会拷贝轮廓outline的transform信息
        """
        if self.inherit_transform:
            return self.outline.local_bounds
        else:
            return self.outline.bounds

    @cached_property
    def bars(self):
        ret = self._create_base_framework()

        if self.inherit_transform:
            ret.transforms = self.outline.transforms.clone()

        return ret

    @cached_property
    def shapes(self):
        return [
            self.bars,
            self.outline
        ]

    def _create_parts(self, axis):
        # 长轴为c，坐标为a，b
        a, b, c = axes = np.r_[axis + 1, axis + 2, axis] % 3  # 三个轴的下标
        ra, rb = self.bar_radius[a], self.bar_radius[b]
        c0, c1 = self.axis_span[c]

        axes = axes[axes]

        for ax, bx in product(self.get_centers(a), self.get_centers(b)):
            origin = np.r_[ax - ra, bx - rb, c0][axes]
            end = np.r_[ax + ra, bx + rb, c1][axes]
            yield Cuboid(origin=origin, end=end)

    def _create_base_framework(self):
        return Composition(
            *self._create_parts(0),
            *self._create_parts(1),
            *self._create_parts(2),
            mix_mode=Composition.Union
        )



from .google_map_source import GoogleMapSource
from .amap_source import AMapSource
from .wayback_source import WaybackSource

from .simple_tile_map_source import SimpleTileMapSource
from .simple_tile_locator import SimpleTileLocator

from .tile_map_source import TileMapSource
from .tile_locator import TileLocator



import psutil

from .executor import Executor, WorkerContext
from .lazy_evaluator import LazyEvaluator
from .utils import traverse_args
from .worker import Worker


class LinearExecutor(Executor):
    def __init__(self, n_units=None):
        """
        在当前进程上构造基于单线程执行器，常用于调试
        :param n_units: 工作单位数
        """
        super().__init__()
        if n_units is None:
            n_units = psutil.cpu_count(logical=False)

        self.workers = [Worker(f'local-{i}', 1) for i in range(n_units)]

    def do_submit(self, func, *args, workers=None, **kwargs):
        # 自动提取所有future
        args, kwargs = traverse_args(args, kwargs,
                                     lambda x: x if not isinstance(x, LazyEvaluator) else self.gather([x])[0])
        return LazyEvaluator(func, *args, **kwargs)

    def progress(self, total=0, **kwargs):
        ret = super().progress(total=total, **kwargs)
        ret.enable_local_lock(False)
        return ret

    def get_workers(self):
        return self.workers

    def is_local(self):
        return True

    def _gather_single(self, future):
        return future.result()

    def get_worker_context(self):
        return LinearWorkerContext(self)

    def receive_events(self, timeout):
        pass

    def check_if_events_thread_are_required(self):
        return False


class LinearWorkerContext(WorkerContext):
    def __init__(self, linear_client: LinearExecutor):
        self.linear_client = linear_client

    def fire(self, event, *args, **kwargs):
        self.linear_client.dispatch_event(event, args, kwargs)



from typing import Sequence

from .array_ops import OperatorDispatcher, ArrayType

Concat = OperatorDispatcher('Concat').aliases({
    ArrayType.numpy: 'concatenate',
    ArrayType.pandas: 'concat',
})


def concat(arrays: Sequence, axis=0):
    """拼接数组

    Parameters
    ----------
    arrays
        数组列表
    axis
        拼接轴向

    Returns
    -------
    ret
        返回数组列表以对应方式拼接的结果
    """
    return Concat.dispatch(type(arrays[0]))(arrays, axis=axis)



from .array_ops import OperatorDispatcher, ArrayType

GetColumn = OperatorDispatcher('GetColumn')


def get_column(arr, ind):
    """从数组获取列

    Parameters
    ----------
    arr
        数组实例
    ind
        列索引或列索引数组 / 切片slice / 列名或列名数组（pandas）

    Returns
    -------
    ret
        返回获取的列，如果输入为切片范围或索引列表，则返回对应矩阵
    """
    return GetColumn.dispatch(type(arr))(arr, ind)


@GetColumn.register(ArrayType.numpy)
def _get_column_np(arr, ind):
    return arr[:, ind]


@GetColumn.register(ArrayType.pandas)
def _get_column_pd(arr, ind):
    if not isinstance(ind, slice):  # slice直接索引会默认索引行，需要排除
        try:
            return arr[ind]  # 判断是否为合法的列key
        except KeyError:
            pass

    return arr.iloc[:, ind]  # 否则认为是列索引



import numpy as np
import tqdm


class QuickUnion:
    def __init__(self, n):
        """快速并查集
        """
        # TODO: 考察是否有必要使用int64作为索引类型
        self.unions = np.arange(n)

    def connect(self, a, b):
        rdst = self.find_root(a)
        rsrc = self.find_root(b)
        self.unions[rsrc] = rdst

    def find_root(self, a):
        root = self.unions[a]
        while root != a:
            a = root
            root = self.unions[a]

        return root

    def collapse(self, verbose=False):
        if verbose:
            span = tqdm.trange(len(self.unions), desc='Collapsing unions')
        else:
            span = range(len(self.unions))

        for i in span:
            a = i
            root = self.unions[a]
            while root != a:
                a = root
                root = self.unions[a]
            self.unions[i] = root



from . import AerialSurvey


class AerialSurveyPoint(AerialSurvey):
    def __init__(self, position, data=None):
        super().__init__(position, data)

    @property
    def point_distances(self):
        return 0

    @property
    def length(self):
        return 0

    @property
    def area(self):
        return 0

    def __len__(self):
        return 1



from __future__ import annotations

from typing import Iterable

from metalpy.utils.regex_pattern import RegexPattern
from .tile_map_source import TileMapSource
from .simple_tile_locator import SimpleTileLocator


class SimpleTileMapSource(TileMapSource):
    def __init__(self, pattern: str | Iterable[str], bottom_left_as_origin=False, regex=False):
        """简单Tile服务器地图源

        Parameters
        ----------
        pattern
            地图源，一般为一个格式化URL，
            例如`https://domain.com/<path>/{z}{x}{y}.png`
            或者`https://domain.com/<path>/{level}/{col}/{row}.png`。
        bottom_left_as_origin
            指示是否以左下方为原点（标准WMTS约定左上方为原点）
        regex
            指示pattern串是否为有穷正则表达式，若为True，会使用RegexPattern将其展开为各个子域名

        Notes
        -----
        若启动regex，{level}, {row}, {y}这些tile坐标占位符会被错误识别为正则的限定次数重复语法，此时需要使用 `\\\\{` 进行转义

        References
        ----------
        metalpy.utils.regex_pattern.RegexPattern: 有穷正则表达式展开算法
        """
        super().__init__()
        self.pattern = pattern
        self.regex = regex
        self.bottom_left_as_origin = bottom_left_as_origin

    def __iter__(self) -> Iterable[SimpleTileLocator]:
        if isinstance(self.pattern, str):
            if self.regex:
                patterns = RegexPattern(self.pattern)
            else:
                patterns = (self.pattern,)
        else:
            patterns = self.pattern

        for pat in patterns:
            yield SimpleTileLocator(pat, bottom_left_as_origin=self.bottom_left_as_origin)



from .texture import Texture, EmptyTexture, TextureSet
from .coefficient import Coefficient

from .ambient import Ambient
from .diffuse import Diffuse
from .specular import Specular
from .alpha import Alpha

from .pbr import PBR, Metallic, Roughness



import numpy as np

from . import Shape3D, Cuboid, BarFramework


class PlateFramework(BarFramework):
    def __init__(self,
                 outline: Shape3D,
                 spec,
                 n_rooms=None,
                 room_spec=None,
                 bounds=None,
                 *,
                 inherit_transform=False
                 ):
        """定义一个规则板结构，其边界由 `outline` 决定，并基于 `outline` 裁剪边界。

        See Also
        --------
        :class:`BarFramework` : 规则框架模型
        """
        super().__init__(
            outline, spec, n_rooms, room_spec, bounds,
            inherit_transform=inherit_transform
        )

    def _create_parts(self, axis):
        # axis为厚度轴，其余两个轴为平面轴
        a, b, c = axes = np.r_[axis, axis + 1, axis + 2] % 3  # 三个轴的下标
        ra = self.bar_radius[a]
        b0, b1 = self.axis_span[b]
        c0, c1 = self.axis_span[c]

        axes = axes[axes]

        for ax in self.get_centers(a):
            origin = np.r_[ax - ra, b0, c0][axes]
            end = np.r_[ax + ra, b1, c1][axes]
            yield Cuboid(origin=origin, end=end)



from metalpy.utils.dhash import dhash
from .mix_modes import MixMode, Mixer, dhashable_mixer
from .object import Object


class Layer:
    def __init__(self, mix_mode: Mixer = MixMode.Override):
        self.objects: list[Object] = []
        self.mix_mode = mix_mode

    @property
    def mixer(self):
        return MixMode.dispatch(self.mix_mode)

    def append(self, obj: Object):
        self.objects.append(obj)

    def clear(self):
        self.objects.clear()

    def __len__(self):
        return len(self.objects)

    def __getitem__(self, item):
        return self.objects[item]

    def __iter__(self):
        for obj in self.objects:
            yield obj

    def __dhash__(self):
        return dhash(*self.objects, dhashable_mixer(self.mix_mode))



from __future__ import annotations

import copy
from collections import namedtuple
from typing import Sequence, Generator, overload, Any

import numpy as np

from metalpy.aero.utils.line_analysis import diff_directions, remove_aux_flights_by_directions
from metalpy.carto.coords import Coordinates
from metalpy.utils.batch import Batch
from metalpy.utils.matplotlib import check_axis
from .aerial_survey import AerialSurvey
from .aerial_survey_line import AerialSurveyLine


class AerialSurveyLines:
    LineSpec = namedtuple('LineSpec', ['direction', 'line_interval'])

    def __init__(self, lines: Sequence[AerialSurveyLine]):
        """构造航空探测测线对象

        Parameters
        ----------
        lines
            测线信息
        """
        self.lines = np.fromiter(lines, dtype=object)

    @property
    def length(self):
        return sum(line.length for line in self)

    def pop_auxiliary(self, direction_threshold: float = 6):
        factors = [line.get_line_spec(check_std=False)[:2] for line in self]
        directions = [np.arctan2(-a, b) for a, b in factors]
        weights = [len(line) for line in self]

        kept_idx = remove_aux_flights_by_directions(directions, weights, stable_threshold=direction_threshold)
        ptr = 0

        kept = []
        popped = []
        for i in range(len(self)):
            line_i = self.lines[i]
            if i != kept_idx[ptr]:
                popped.append(line_i)
            else:
                kept.append(line_i)
                ptr += 1

                if ptr >= len(kept_idx):
                    # 剩下的全部被视为辅助航线，加入排除列表
                    for j in range(i + 1, len(self)):
                        popped.append(self.lines[j])
                    break

        self.lines = kept
        return AerialSurveyLines(popped)

    def remove_auxiliary(self, direction_threshold: float = 6, inplace=False):
        if inplace:
            ret = self
        else:
            ret = copy.copy(self)

        ret.pop_auxiliary(direction_threshold=direction_threshold)
        return ret

    def merge_neighbors(self, inplace=False):
        directions, intercepts, cos_thetas = self.get_detailed_line_specs()

        di = abs(np.diff(intercepts)) * cos_thetas[:-1]
        criteria = np.percentile(di, 40) / 2

        merged_lines = [[self.lines[0]]]
        for i, merged in enumerate(di < criteria):
            line = self.lines[i + 1]
            if merged:
                merged_lines[-1].append(line)
            else:
                merged_lines.append([line])

        merged_lines = [AerialSurveyLine.concat(lines) for lines in merged_lines]

        if inplace:
            self.lines = merged_lines
            ret = self
        else:
            ret = AerialSurveyLines(merged_lines)

        return ret

    @overload
    def __getitem__(self, item: int) -> AerialSurveyLine: ...

    @overload
    def __getitem__(self, item: Any) -> AerialSurveyLines: ...

    def __getitem__(self, item) -> AerialSurveyLine | AerialSurveyLines:
        ret = self.lines[item]
        if np.ndim(ret) == 0:
            return ret
        else:
            return AerialSurveyLines(ret)

    def __len__(self):
        return len(self.lines)

    def __iter__(self) -> Generator[AerialSurveyLine]:
        yield from self.lines

    def merge(self):
        return AerialSurveyLine.concat(self.lines)

    def to_survey(self):
        ret = self.merge()
        return AerialSurvey(ret.position, ret.data)

    def get_line_specs(self, group_direction_tol=10):
        directions, intercepts, cos_thetas = self.get_detailed_line_specs()
        directions = np.rad2deg(directions)

        splits = np.where(abs(diff_directions(directions)) > group_direction_tol)[0] + 1
        directions = [ds.mean() for ds in np.array_split(directions, splits)]
        cos_thetas = [ct.mean() for ct in np.array_split(cos_thetas, splits)]
        intervals = [
            np.abs(np.diff(ints) * ct).mean()
            for ints, ct in zip(np.array_split(intercepts, splits), cos_thetas)
        ]

        return [
            AerialSurveyLines.LineSpec(d, i)
            for d, i in zip(directions, intervals)
        ]

    def get_detailed_line_specs(self):
        """获取详细的航线分析信息

        Returns
        -------
        directions, intercepts, cos_thetas
            每条航线的朝向（-pi/2~pi/2），每条航线到最近坐标轴的截距，每条航线到最近坐标轴的余弦角
        """
        lines = Batch.of(self.lines)
        line_specs = np.asarray(lines.get_line_spec())
        centers = np.asarray(lines.center)
        rx, ry = centers.mean(axis=0)  # 参考原点

        directions = []
        intercepts = []
        slopes = []
        for dx, dy, intercept in line_specs:
            if dy == -1:
                directions.append(np.arctan(dx))
                intercepts.append(dx * rx + intercept)
                slopes.append(dx)
            else:
                # 注意其定义域为[(1/2) * pi, (3/2) * pi]
                # 需要进行转换
                directions.append(np.pi / 2 - np.arctan(dy))
                intercepts.append(dy * ry + intercept)
                slopes.append(dy)

        directions = np.asarray(directions)
        intercepts = np.asarray(intercepts)
        slopes = np.asarray(slopes)
        directions[directions > np.pi / 2] -= np.pi

        cos_thetas = 1 / np.sqrt(1 + slopes ** 2)

        return directions, intercepts, cos_thetas

    def plot(self, ax=None, show=True, fontsize=10, **kwargs):
        with check_axis(ax, show=show, figsize=(5, 5)) as ax:
            for i, line in enumerate(self.lines):
                pos = np.asarray(line.position)
                ax.plot(*pos.T, **kwargs)
                ax.text(*Coordinates(pos).bounds.center, str(i), fontsize=fontsize)

    def __copy__(self):
        return AerialSurveyLines(self.lines)



import functools

from metalpy.utils.object_path import get_full_qualified_path, reassign_object_name


class XX:
    class YY:
        def zz(self):
            pass


class Target:
    pass


def test_full_qualname():
    assert get_full_qualified_path(XX.YY.zz).endswith('XX.YY.zz')

    XX.YY.zz.__module__ = __name__
    assert get_full_qualified_path(XX.YY.zz) == f'{__name__}:XX.YY.zz'

    XX.YY.zz.__name__ = 'ZZ'
    assert get_full_qualified_path(XX.YY.zz) != f'{__name__}:XX.YY.ZZ'  # 只改了__name__但没有改__qualname__

    reassign_object_name(XX.YY.zz, new_name='ZZ')
    assert get_full_qualified_path(XX.YY.zz) == f'{__name__}:XX.YY.ZZ'

    mock = XX()
    functools.wraps(Target)(mock)
    assert get_full_qualified_path(mock) == get_full_qualified_path(Target)



def drop_by_indices(df, indices, ext=0):
    import pandas as pd

    dropped = df.copy()
    length = max(dropped.index)
    for idx in indices:
        start, end = min(idx), max(idx)
        dropped.drop(index=idx, inplace=True, errors='ignore')
        if ext > 0:
            dropped.drop(index=pd.RangeIndex(max(min(idx) - ext, 0), start), inplace=True, errors='ignore')
            dropped.drop(index=pd.RangeIndex(end, min(max(idx) + ext, length)), inplace=True, errors='ignore')

    return dropped



import warnings

from .simple_tile_map_source import SimpleTileMapSource


class GoogleMapSource(SimpleTileMapSource):
    LangDefault = None
    LangChs = 'zh-CN'
    OffsetChs = 'cn'

    def __init__(self,
                 satellite=False,
                 labels=True,
                 arterial=None,
                 terrain=False,
                 labels_only=False,
                 terrain_only=False,
                 lang=None,
                 offset=None,
                 scale=None
                 ):
        """谷歌地图数据源

        Parameters
        ----------
        satellite
            是否使用卫星底图
        labels
            是否包含标签
        arterial
            是否包含主干道（不确定），对应选项 lyrs=m，只在带标签常规地图模式下有效，默认为True
            （not satellite and labels）
        terrain
            是否包含地形图，只在带标签常规地图模式下有效
            （not satellite and labels）
        labels_only
            下载只有标签的底图
        terrain_only
            下载只有地形图的底图
        lang
            地图渲染格式以及标注语言，即hl字段
        offset
            地图偏移，即gl字段，如果输入坐标为GCJ02坐标系则应该可以通过`offset=OffsetChs`校正
        """
        if terrain and arterial:
            warnings.warn('Terrain and arterial layer cannot exists at same time.')
        if terrain:
            if satellite or not labels:
                raise ValueError('Terrain layer supports only default map with labels.')
        if arterial:
            if satellite or not labels:
                warnings.warn('Arterial layer supports only default map with labels.')

        if arterial is None:
            arterial = True

        components = [r'http://mt[0123].google.com/vt/x=\{x\}&y=\{y\}&z=\{z\}&s=Gali']

        layers = 'm'
        if labels_only:
            layers = 'h'
        elif terrain_only:
            layers = 't'
        elif satellite:
            if labels:
                layers = 'y'
            else:
                layers = 's'
        elif not satellite:
            if labels:
                if terrain:
                    layers = 'p'
                elif arterial:
                    layers = 'm'
                else:
                    layers = 'r'
            else:
                layers = 't'

        components.append(f'lyrs={layers}')

        if lang is not None:
            components.append(f'hl={lang}')
        if offset is not None:
            components.append(f'gl={offset}')

        if scale is not None:
            components.append(f'scale={scale}')

        super().__init__(
            '&'.join(components),
            regex=True
        )



import os

import numpy as np
from SimPEG.potential_fields.magnetics import Simulation3DIntegral
from discretize.utils import mkvc

from metalpy.scab.progressed import Progress

from .tied_simulation3d_integral import TaichiSimulation3DIntegral, Receiver
from ...taichi_kernel_base import TiedMixin, tied_profile


class TiedSimulation3DIntegralMixin(TiedMixin):
    def __init__(self, this, **kwargs):
        """
        Notes
        -----
        SimPEG原版的磁正演过程会忽略mesh的orientation属性，直接进行正演。
        TiedSimulation3DIntegralMixin通过逆向旋转所有相关量来实现了网格旋转。
        因此如果mesh.orientation不为单位矩阵（即存在旋转），则行为会与SimPEG不一致。
        """
        super().__init__(this, **kwargs)

    @tied_profile
    def linear_operator(this, self: Simulation3DIntegral):
        n_cells = self.nC

        if getattr(self, "model_type", None) == "vector":
            n_cells *= 3
        if self.store_sensitivities == "disk":
            sens_name = os.path.join(self.sensitivity_path, "sensitivity.npy")
            if os.path.exists(sens_name):
                # do not pull array completely into ram, just need to check the size
                kernel = np.load(sens_name, mmap_mode="r")
                if kernel.shape == (self.survey.nD, n_cells):
                    print(f"Found sensitivity file at {sens_name} with expected shape")
                    kernel = np.asarray(kernel)
                    return kernel

        forward_only = self.store_sensitivities == "forward_only"
        if forward_only:
            model = self.chi  # chi = mapping * model
        else:
            model = None

        # 计算网格在三个方向的边界位置
        cell_centers = self.mesh.cell_centers[self.ind_active]
        h_gridded = self.mesh.h_gridded[self.ind_active]
        bsw = cell_centers - h_gridded / 2.0
        tne = cell_centers + h_gridded / 2.0

        xn1, xn2 = bsw[:, 0], tne[:, 0]
        yn1, yn2 = bsw[:, 1], tne[:, 1]
        zn1, zn2 = bsw[:, 2], tne[:, 2]

        xn = np.c_[mkvc(xn1), mkvc(xn2)]
        yn = np.c_[mkvc(yn1), mkvc(yn2)]
        zn = np.c_[mkvc(zn1), mkvc(zn2)]

        base_cell_sizes = np.r_[
            self.mesh.h[0].min(),
            self.mesh.h[1].min(),
            self.mesh.h[2].min(),
        ]

        if self.model_type == 'scalar':
            magnetization = self.M  # (nC, 3)的矩阵
        else:
            # 输入向量，TaichiSimulation3DIntegral内部会自动转换为(nC, 3)的矩阵
            # 向量模式下（model_type='vector'），dpred输入为三轴等效磁化率，乘以地磁场幅值得到磁化矩阵（地磁场方向此时不参与计算）
            # 与Simulation3DIntegral.evaluate_integral中`if self.model_type == "vector"`分支行为一致
            magnetization = self.survey.source_field.amplitude

        if self.model_type == 'scalar':
            model_type = TaichiSimulation3DIntegral.MType_Scalar
        else:
            model_type = TaichiSimulation3DIntegral.MType_Vector
            if forward_only:
                model = model.reshape(3, -1).T

        receivers = [Receiver(rx.locations, rx.components) for rx in self.survey.source_field.receiver_list]

        progress = self.mixins.get(Progress)

        kernel = TaichiSimulation3DIntegral(
            receivers=receivers,
            xn=xn, yn=yn, zn=zn,
            base_cell_sizes=base_cell_sizes,
            model_type=model_type,
            magnetization=magnetization,
            row_stype=TaichiSimulation3DIntegral.Layout_SoA,
            col_stype=TaichiSimulation3DIntegral.Layout_AoS,
            tmi_projection=self.tmi_projection,
        ).dpred(model, progress=progress)

        if self.store_sensitivities == "forward_only":
            kernel = kernel.ravel()

        if self.store_sensitivities == "disk":
            print(f"writing sensitivity to {sens_name}")
            os.makedirs(self.sensitivity_path, exist_ok=True)
            np.save(sens_name, kernel)

        return kernel



from typing import Union

import numpy as np


def get_prolate_spheroid_demag_factor(eps, polar_axis=2):
    """
    获取长旋转椭球体的退磁系数（a, a, c, c > a）

    Parameters
    ----------
    eps : float
        椭球长短轴比 (c / a)

    polar_axis : int
        长轴所在轴方向，0、1、2代表x、y、z，默认为2，即z轴

    Returns
    -------
    ret
        长旋转椭球体的退磁系数
    """
    ratio = eps
    ratio2m1 = ratio * ratio - 1
    sratio2m1 = np.sqrt(ratio2m1)
    Na = (ratio / sratio2m1 * np.log(ratio + sratio2m1) - 1) / ratio2m1
    Nc = (1 - Na) / 2
    N = np.asarray([Nc, Nc, Nc])
    N[polar_axis] = Na

    return N


def get_demag_apparent_k(k: Union[float, np.ndarray], N):
    """
    计算退磁作用下的视磁化率

    Parameters
    ----------
    k : float | array-like(3,) | array-like(n, 3)
        均匀介质磁化率
    N : array-like(3,)
        三个方向上的退磁系数

    Returns
    -------
    ret : array(n, 3)
        视磁化率
    """
    N = np.asarray(N)
    k = np.asarray(k)

    if k.ndim < 2:
        k = k[..., None]

    return k / (1 + k * N)



from . import Texture


class Ambient(Texture):
    def apply(self, actor):
        actor.prop.ambient_color = self.texture
        # 系数如果取1会过曝，难道是PyVista的plotter默认亮度太高了？
        actor.prop.ambient = 0.1



from abc import ABC, abstractmethod

from metalpy.utils.dhash import dhash


class Transform(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def transform(self, mesh):
        return mesh

    @abstractmethod
    def inverse_transform(self, mesh):
        return mesh

    @abstractmethod
    def clone(self):
        return Transform()

    def __hash__(self):
        return dhash(self).digest()

    @abstractmethod
    def __dhash__(self):
        raise NotImplementedError()



from scipy.spatial.transform import Rotation as R

from metalpy.utils.dhash import dhash
from .transform import Transform


class Rotation(Transform):
    def __init__(self, y, a, b, degrees=False, seq='xyz'):
        super().__init__()
        self.params = (y, a, b, degrees, seq)
        # TODO: 解决Rotation精度问题
        self.rot = R.from_euler(seq, [y, a, b], degrees=degrees)

    def transform(self, mesh):
        return self.rot.apply(mesh)

    def inverse_transform(self, mesh):
        return self.rot.apply(mesh, inverse=True)

    def clone(self):
        return Rotation(*self.params)

    def __dhash__(self):
        return dhash(*self.params[:-1], self.params[-1])



import importlib
import os
import shutil

import sys
import uuid
import zipfile

from dask.utils import tmpfile
from distributed import WorkerPlugin, NannyPlugin
from distributed.utils import logger

from metalpy.utils.type import ensure_as_iterable


def configure_dask_client(client, extra_paths=None, excludes=None):
    class UploadModules(NannyPlugin):
        def __init__(self, identifier=None):
            self.modules = []

            if identifier is None:
                self.identifier = str(uuid.uuid4())
            else:
                self.identifier = str(identifier)

        def add(self, module_name, path):
            self.modules.append((module_name, path))

        def compact(self):
            with tmpfile() as fn:
                with zipfile.ZipFile(fn, "w") as z:
                    for module_name, path in self.modules:
                        module_path = module_name.replace(".", "/")
                        path = os.path.abspath(path).replace("\\", "/")
                        index = path.find(module_path)
                        archive_name = path[index:]

                        z.write(path, archive_name)

                with open(fn, "rb") as f:
                    self.data = f.read()

        def get_reloader(self):
            return ReloadModules(self.modules)

        async def setup(self, nanny):
            target = nanny

            fn = os.path.join(target.local_directory, f"tmp-{self.identifier}.zip")
            if not os.path.exists(fn):
                with open(fn, "wb") as f:
                    f.write(self.data)

                import zipfile

                base_dir = target.local_directory
                if 'dask-worker-space' not in target.local_directory:
                    # 在不同配置下nanny的工作目录可能在dask-worker-space的更上一级目录
                    # 具体来说是如果通过 --local-directory 指定了worker的工作路径，nanny的工作路径就会变为
                    # %working_dir% 而不是之前的 %working_dir%/dask-worker-space
                    # 此时应该切进 dask-worker-space 以保持一致
                    base_dir = os.path.join(base_dir, 'dask-worker-space')

                check_directory_flag = False
                for content in os.listdir(base_dir):
                    if 'worker-' in content:
                        check_directory_flag = True
                        break
                if not check_directory_flag:
                    logger.warning('Wrong nanny working directory: %s', base_dir)

                path_to_scripts = os.path.join(base_dir, "scripts")

                if os.path.exists(path_to_scripts):
                    shutil.rmtree(path_to_scripts)

                with zipfile.ZipFile(fn) as z:
                    z.extractall(path=path_to_scripts)

                os.remove(fn)

    class ReloadModules(WorkerPlugin):
        def __init__(self, modules):
            self.modules = modules

        async def setup(self, worker=None):
            target = worker
            base_dir = os.path.abspath(os.path.join(target.local_directory, '../..'))
            if 'dask-worker-space' not in base_dir:
                # 同上，如果nanny的工作路径不在dask-worker-space下，那么worker的工作路径似乎也不会在dask-worker-space下
                base_dir = os.path.join(base_dir, 'dask-worker-space')

            path_to_scripts = os.path.join(base_dir, "scripts")

            print(path_to_scripts)

            importlib.invalidate_caches()
            sys.path.insert(0, path_to_scripts)

            # import modules in the scripts folder
            for module_name, path in self.modules:
                logger.info("Reloading module %s from .py file", module_name)
                importlib.reload(importlib.import_module(module_name))

            sys.path.remove(path_to_scripts)

    extra_paths = ensure_as_iterable(extra_paths, excludes=str)
    excludes = ensure_as_iterable(excludes, excludes=str)

    paths = [os.path.abspath(p) for p in extra_paths]

    current_dir = os.path.dirname(os.path.abspath(__file__))
    current_module_location = __name__.split('.')

    while os.path.split(current_dir)[1] in current_module_location:
        current_dir = os.path.split(current_dir)[0]

    paths.append(current_dir)
    paths.append(os.path.abspath('./'))

    # 将脚本文件上传到服务器
    modules = list(sys.modules.items())
    plugin = UploadModules(uuid.uuid4())

    for name, module in modules:
        try:
            file_path = module.__spec__.origin
        except AttributeError:
            continue

        if file_path is None:
            continue

        required = False
        for path in paths:
            if path not in os.path.dirname(file_path):
                continue
            else:
                required = True
                break

        if not required:
            continue

        for exclude in excludes:
            if hasattr(exclude, '__call__'):
                required = not exclude(name)
            else:
                required = exclude not in name

            if not required:
                break

        if not required:
            continue

        plugin.add(name, file_path)

    plugin.compact()

    client.register_worker_plugin(plugin, nanny=True, name='UploadUserModules')
    client.register_worker_plugin(plugin.get_reloader(), name='ReloadUserModules')
    client.unregister_worker_plugin(name='ReloadUserModules')
    client.unregister_worker_plugin(nanny=True, name='UploadUserModules')



import sys

import numpy as np

from metalpy.utils.bounds import Bounds
from metalpy.utils.dhash import dhash
from metalpy.utils.string import parse_axes_labels, parse_labels
from . import InfiniteShape


class SurfaceTopo(InfiniteShape):
    def __init__(self, loc, alt=None, loc_axes=None, alt_axis=None, inverse=False, sample=None):
        """定义一个地形几何体

        Parameters
        ----------
        loc
            地形位置坐标
        alt
            地形高度，若未指定，则取`loc`的最后一列为高度
        loc_axes
            地形位置所属的坐标轴，默认为 `xy`
        alt_axis
            地形的外法线方向，指向地面以上，例如
            `+z` / `z` / `2` 代表垂直xy平面向+z方向，则-z方向部分为 `地下`，
            `-y` / `-1` 代表垂直xz平面向-y方向，则+y方向部分为 `地下`
        inverse
            翻转地形的外法线方向。
            注：如果 `alt_axis` 为 `-z` 则会变为 `+z`
        sample
            对曲面上的点进行采样的采样间距，等价于 loc[::sample], alt[::sample]
        """
        super().__init__()

        if loc_axes is not None:
            loc_axes = parse_axes_labels(loc_axes, max_length=2)
            loc_axes = list(set(loc_axes))

        if alt_axis is not None:
            self.alt_axis, self.inverse = parse_axis_name(alt_axis)
            if loc_axes is None:
                loc_axes = list({0, 1, 2}.difference(self.alt_axis))
            self.loc_axes = loc_axes
        elif loc_axes is not None:
            self.loc_axes = loc_axes
            assert self.surface_dim == 2, 'Only 3D surface can have `alt_axis` inferred from `loc_axes`.'
            self.alt_axis, self.inverse = outer_product(loc_axes)
        else:
            raise RuntimeError(
                'Either of `loc_axes` or `alt_axis` must be provided.'
            )

        if inverse:
            self.inverse = not self.inverse

        loc = np.asarray(loc)
        if alt is None:
            alt = loc[:, self.alt_axis]
            loc = loc[:, self.loc_axes]
        else:
            alt = np.asarray(alt)
            assert len(loc) == len(alt), '`loc` and `alt` must have same length.'

        assert self.surface_dim == loc.shape[1], '`loc_axes` must have same dimensions with locations.'

        if sample is not None:
            loc = loc[::sample]
            alt = alt[::sample]

        self.loc = loc
        self.alt = alt

    @property
    def surface_dim(self):
        return len(self.loc_axes)

    @property
    def n_pts(self):
        return len(self.loc)

    @property
    def n_dims(self):
        return self.surface_dim + 1

    def track(self, points):
        from scipy.interpolate import LinearNDInterpolator

        fun_interp = LinearNDInterpolator(self.loc, self.alt)
        return fun_interp(points)

    @staticmethod
    def xy2z(loc, alt=None, inverse=False, sample=None):
        return SurfaceTopo.ab2c(loc, alt=alt, loc_axes='xy', alt_axis='z', inverse=inverse, sample=sample)

    @staticmethod
    def yz2x(loc, alt=None, inverse=False, sample=None):
        return SurfaceTopo.ab2c(loc, alt=alt, loc_axes='yz', alt_axis='x', inverse=inverse, sample=sample)

    @staticmethod
    def zx2y(loc, alt=None, inverse=False, sample=None):
        return SurfaceTopo.ab2c(loc, alt=alt, loc_axes='zx', alt_axis='y', inverse=inverse, sample=sample)

    @staticmethod
    def x2y(loc, alt=None, inverse=False, sample=None):
        return SurfaceTopo.ab2c(loc, alt=alt, loc_axes='x', alt_axis='y', inverse=inverse, sample=sample)

    @staticmethod
    def y2z(loc, alt=None, inverse=False, sample=None):
        return SurfaceTopo.ab2c(loc, alt=alt, loc_axes='y', alt_axis='z', inverse=inverse, sample=sample)

    @staticmethod
    def z2x(loc, alt=None, inverse=False, sample=None):
        return SurfaceTopo.ab2c(loc, alt=alt, loc_axes='z', alt_axis='x', inverse=inverse, sample=sample)

    @staticmethod
    def ab2c(loc, alt=None, loc_axes='xy', alt_axis='z', inverse=False, sample=None):
        pv = sys.modules.get('pyvista', None)
        if pv is not None:
            if isinstance(loc, pv.DataSet):
                return SurfaceTopo.from_polydata(
                    loc,
                    loc_axes=loc_axes,
                    alt_axis=alt_axis,
                    inverse=inverse,
                    sample=sample
                )
        return SurfaceTopo(
            loc, alt=alt,
            loc_axes=loc_axes,
            alt_axis=alt_axis,
            inverse=inverse,
            sample=sample
        )

    @staticmethod
    def from_polydata(poly, loc_axes=None, alt_axis=None, inverse=False, sample=None):
        from metalpy.utils.pyvista_dataset_wrappers.universal_dataset import UniversalDataSet
        points = UniversalDataSet(poly).points

        if loc_axes is None and alt_axis is None:
            maxs, mins = points.max(axis=0), points.min(axis=0)
            scaled = (points - mins) / (maxs - mins)
            # 协方差最接近0，代表其在对应平面上分布最均匀，因此大概率为坐标平面
            cov = np.cov(scaled, rowvar=False).ravel()[[1, 2, 5]]
            loc_axes = [[0, 1], [1, 2], [2, 0]][np.argmin(np.abs(cov))]

        return SurfaceTopo(
            points,
            loc_axes=loc_axes,
            alt_axis=alt_axis,
            inverse=inverse,
            sample=sample
        )

    def do_place(self, mesh_cell_centers, progress):
        cell_locs = mesh_cell_centers[:, self.loc_axes]
        cell_alts = mesh_cell_centers[:, self.alt_axis]

        if not self.inverse:
            return cell_alts < self.track(cell_locs)
        else:
            return cell_alts > self.track(cell_locs)

    def __dhash__(self):
        return dhash(
            super().__dhash__(),
            self.loc,
            self.alt,
            self.loc_axes,
            self.alt_axis,
            self.inverse,
        )

    def do_clone(self, deep=True):
        return SurfaceTopo(loc=self.loc, alt=self.alt, loc_axes=self.loc_axes, alt_axis=self.alt_axis,
                           inverse=self.inverse)

    @property
    def local_bounds(self):
        ret = Bounds.unbounded(self.n_dims)

        for idx, a in enumerate(self.loc_axes):
            c = self.loc[:, idx]
            ret.set(a, min=c.min(), max=c.max())

        if self.inverse:
            ret.set(self.alt_axis, min=self.alt.min())
        else:
            ret.set(self.alt_axis, max=self.alt.max())

        return ret

    def to_local_polydata(self):
        import pyvista as pv

        if self.surface_dim < 2:
            return None  # TODO: Scene添加接口支持获取场景信息来构造polydata

        pts = np.empty(
            (self.n_pts, self.surface_dim + 1),
            dtype=np.result_type(self.loc, self.alt)
        )
        pts[:, self.loc_axes] = self.loc
        pts[:, self.alt_axis] = self.alt

        from scipy.spatial import Delaunay
        delaunay = Delaunay(self.loc)
        faces = delaunay.simplices
        _3 = np.full(len(faces), 3)

        return pv.PolyData(pts, np.c_[_3, faces].ravel())


def axis_direction(axis, inverse=False):
    if not inverse:
        return f'+{axis}'
    else:
        return f'-{axis}'


def parse_axis_name(axis):
    if isinstance(axis, int):
        if axis > 0:
            inverse = False
        else:
            inverse = True
        axis = abs(axis)
    elif len(axis) == 1:
        inverse = False
    elif len(axis) == 2:
        inverse = [False, True][parse_labels(axis[0], accepts='+-', length=1)[0]]
        axis = axis[1]
    else:
        raise RuntimeError(
            f'`axis` must be axis name like `+x`, `-y` or `z`(alias to `+z`),'
            f' got {axis}.'
        )
    axis = parse_axes_labels(axis, length=1)[0]

    return axis, inverse


def outer_product(axes):
    """根据右手定则确定第三轴方向，认为 `axes` 分别指定了两个坐标轴的正方向

    则 `xy` , `yz` , `zx` 分别对应 `+z` , `+x` , `+y`

    Parameters
    ----------
    axes
        参与计算的两个坐标轴，必须为 0, 1, 2 中的两个

    Returns
    -------
    axis_with_sign
        返回外积下的第三轴方向
    """
    axes_set = set(axes)
    assert len(axes_set) == 2 and len(axes_set.difference({0, 1, 2})) == 0

    axis = list({0, 1, 2}.difference(axes))[0]
    inverse = (axes[1] - axes[0]) % 3 == 2

    return axis, inverse
